from copy import deepcopy
from math import log
import os
import subprocess
import urllib
from dataclasses import dataclass, field

import networkx as nx
import openai
from loguru import logger
from openai.types.beta.thread import Thread
from openai.types.beta.threads.run import Run

from sweepai.config.client import SweepConfig
from sweepai.core.chat import ChatGPT
from sweepai.core.entities import Message, Snippet
from sweepai.logn.cache import file_cache
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.convert_openai_anthropic import AnthropicFunctionCall, mock_function_calls_to_string
from sweepai.utils.github_utils import ClonedRepo
from sweepai.utils.modify_utils import post_process_rg_output
from sweepai.utils.openai_listwise_reranker import listwise_rerank_snippets
from sweepai.utils.progress import AssistantConversation, TicketProgress
from sweepai.utils.tree_utils import DirectoryTree

ASSISTANT_MAX_CHARS = 4096 * 4 * 0.95  # ~95% of 4k tokens
NUM_SNIPPETS_TO_SHOW_AT_START = 15
MAX_REFLECTIONS = 1
MAX_ITERATIONS = 25
NUM_ROLLOUTS = 1 # dev speed
SCORE_THRESHOLD = 8 # good score
STOP_AFTER_SCORE_THRESHOLD_IDX = 0 # stop after the first good score and past this index
MAX_PARALLEL_FUNCTION_CALLS = 1
NUM_BAD_FUNCTION_CALLS = 5

# TODO:
# - Add self-evaluation / chain-of-verification

anthropic_function_calls = """<tool_description>
<tool_name>code_search</tool_name>
<description>
Passes the code_entity into ripgrep to search the entire codebase and return a list of files and line numbers where it appears. Useful for finding definitions, usages, and references to types, classes, functions, and other entities that may be relevant. Review the search results using `view_files` to determine relevance and discover new files to explore.
</description>
<parameters>
<parameter>
<name>analysis</name>
<type>string</type>
<description>Explain what new information you expect to discover from this search and why it's needed to get to the root of the issue. Focus on unknowns rather than already stored information.</description>
</parameter>
<parameter>
<name>code_entity</name>
<type>string</type>
<description>
The code entity to search for. This must be a distinctive name, not a generic term. For functions, search for the definition syntax, e.g. 'def foo' in Python or 'function bar' or 'const bar' in JavaScript. Trace dependencies of critical functions/classes, follow imports to find definitions, and explore how key entities are used across the codebase.
</description>
</parameter>
</parameters>
</tool_description>

<tool_description>
<tool_name>view_files</tool_name>
<description>
Retrieves the contents of the specified file(s). After viewing new files, use `code_search` on relevant entities to continue discovering potentially relevant files. You may view three files per tool call. Prioritize viewing new files over ones that are already stored.
</description>
<parameters>
<parameter>
<name>analysis</name>
<type>string</type>
<description>Explain what new information viewing these files will provide and why it's necessary to resolve the issue. Avoid restating already known information.</description>
</parameter>
<parameter>
<name>first_file_path</name>
<type>string</type>
<description>The path of a new file to view.</description>
</parameter>
<parameter>
<name>second_file_path</name>
<type>string</type>
<description>The path of another new file to view (optional).</description>
</parameter>
<parameter>
<name>third_file_path</name>
<type>string</type>
<description>The path of a third new file to view (optional).</description>
</parameter>
</parameters>
</tool_description>

<tool_description>
<tool_name>store_file</tool_name>
<description>
Adds a newly discovered file that provides important context or may need modifications to the list of stored files. You may only store one new file per tool call. Avoid storing files that have already been added.
</description>
<parameters>
<parameter>
<name>analysis</name>
<type>string</type>
<description>Explain what new information this file provides, why it's important for understanding and resolving the issue, and what potentially needs to be modified. Include a brief supporting code excerpt.</description>
</parameter>
<parameter>
<name>file_path</name>
<type>string</type>
<description>The path of the newly discovered relevant file to store.</description>
</parameter>
</parameters>
</tool_description>

You MUST call the tools using this exact XML format:

<function_call>
<invoke>
<tool_name>$TOOL_NAME</tool_name>
<parameters>
<$PARAMETER_NAME>$PARAMETER_VALUE</$PARAMETER_NAME>
...
</parameters>
</invoke>
</function_call>

Here is an example illustrating a complex code search to discover new relevant information:

<example>
<function_call>
<invoke>
<tool_name>code_search</tool_name>
<parameters>
<analysis>The get_user_by_id method likely queries from a User model or database table. I need to search for references to "User" to find where and how user records are defined, queried and filtered in order to determine what changes are needed to support excluding deleted users from the get_user_by_id results.</analysis>
<code_entity>User</code_entity>
</parameters>
</invoke>
</function_call>
</example>

Remember, your goal is to discover and store ALL files that are relevant to solving the issue. Perform targeted searches to uncover new information, view new files to understand the codebase, and avoid re-analyzing already stored files."""

sys_prompt = """You are a brilliant engineer assigned to solve the following GitHub issue. Your task is to search through the codebase and locate ALL files that are RELEVANT to resolving the issue. A file is considered RELEVANT if it provides important context or may need to be modified as part of the solution.

You will begin with a small set of stored relevant files. However, it is critical that you identify every additional relevant file by exhaustively searching the codebase. Your goal is to generate an extremely comprehensive list of files for an intern engineer who is completely unfamiliar with the codebase. Prioritize finding all relevant files over perfect precision - it's better to include a few extra files than to miss a key one.

To accomplish this, you will iteratively search for and view new files to gather all the necessary information. Follow these steps:

1. Perform targeted code searches to find definitions, usages, and references for ALL unknown variables, classes, attributes, functions and other entities that may be relevant based on the currently stored files and issue description. Be creative and think critically about what to search for to get to the root of the issue. 

2. View new files from the search results that seem relevant. Avoid viewing files that are already stored, and instead focus on discovering new information.

3. Store additional files that provide important context or may need changes based on the search results, viewed files, and issue description. 

Repeat steps 1-3, searching and exploring the codebase exhaustively until you are confident you have found all relevant files. Prioritize discovering new information over re-analyzing what is already known.

Here are the tools at your disposal:
""" + anthropic_function_calls

unformatted_user_prompt = """\
## Stored Files
DO NOT CALL THE STORE OR VIEW TOOLS ON THEM AGAIN AS THEY HAVE ALREADY BEEN STORED.
<stored_files>
{snippets_in_repo}
</stored_files>

{import_tree_prompt}
## User Request
<user_request>
{query}
<user_request>"""

PLAN_SUBMITTED_MESSAGE = "SUCCESS: Report and plan submitted."

def escape_ripgrep(text):
    # Special characters to escape
    special_chars = ["(", "{"]
    for s in special_chars:
        text = text.replace(s, "\\" + s)
    return text

@staticmethod
def can_add_snippet(snippet: Snippet, current_snippets: list[Snippet]):
    return (
        len(snippet.xml) + sum([len(snippet.xml) for snippet in current_snippets])
        <= ASSISTANT_MAX_CHARS
    )


@dataclass
class RepoContextManager:
    dir_obj: DirectoryTree
    current_top_tree: str
    snippets: list[Snippet]
    snippet_scores: dict[str, float]
    cloned_repo: ClonedRepo
    current_top_snippets: list[Snippet] = field(default_factory=list)
    read_only_snippets: list[Snippet] = field(default_factory=list)
    issue_report_and_plan: str = ""
    import_trees: str = ""
    relevant_file_paths: list[str] = field(
        default_factory=list
    )  # a list of file paths that appear in the user query

    @property
    def top_snippet_paths(self):
        return [snippet.file_path for snippet in self.current_top_snippets]

    @property
    def relevant_read_only_snippet_paths(self):
        return [snippet.file_path for snippet in self.read_only_snippets]

    def expand_all_directories(self, directories_to_expand: list[str]):
        self.dir_obj.expand_directory(directories_to_expand)

    def is_path_valid(self, path: str, directory: bool = False):
        if directory:
            return any(snippet.file_path.startswith(path) for snippet in self.snippets)
        return any(snippet.file_path == path for snippet in self.snippets)

    def format_context(
        self,
        unformatted_user_prompt: str,
        query: str,
    ):
        files_in_repo_str = ""
        stored_files = set()
        for idx, snippet in enumerate(list(dict.fromkeys(self.current_top_snippets))[:NUM_SNIPPETS_TO_SHOW_AT_START]):
            if snippet.file_path in stored_files:
                continue
            stored_files.add(snippet.file_path)
            snippet_str = \
f'''
<stored_file index="{idx + 1}">
<file_path>{snippet.file_path}</file_path>
<source>
{snippet.content}
</source>
</stored_file>
'''
            files_in_repo_str += snippet_str
        repo_tree = str(self.dir_obj)
        import_tree_prompt = """
## Import trees for code files in the user request
<import_trees>
{import_trees}
</import_trees>
"""
        import_tree_prompt = (
            import_tree_prompt.format(import_trees=self.import_trees.strip("\n"))
            if self.import_trees
            else ""
        )
        user_prompt = unformatted_user_prompt.format(
            query=query,
            snippets_in_repo=files_in_repo_str,
            repo_tree=repo_tree,
            import_tree_prompt=import_tree_prompt,
            file_paths_in_query=", ".join(self.relevant_file_paths),
        )
        return user_prompt

    def get_highest_scoring_snippet(self, file_path: str) -> Snippet:
        def snippet_key(snippet):
            return snippet.denotation

        filtered_snippets = [
            snippet
            for snippet in self.snippets
            if snippet.file_path == file_path
            and snippet not in self.current_top_snippets
        ]
        if not filtered_snippets:
            return None
        highest_scoring_snippet = max(
            filtered_snippets,
            key=lambda snippet: (
                self.snippet_scores[snippet_key(snippet)]
                if snippet_key(snippet) in self.snippet_scores
                else 0
            ),
        )
        return highest_scoring_snippet

    def add_snippets(self, snippets_to_add: list[Snippet]):
        # self.dir_obj.add_file_paths([snippet.file_path for snippet in snippets_to_add])
        for snippet in snippets_to_add:
            self.current_top_snippets.append(snippet)

    # does the same thing as add_snippets but adds it to the beginning of the list
    def boost_snippets_to_top(self, snippets_to_boost: list[Snippet]):
        # self.dir_obj.add_file_paths([snippet.file_path for snippet in snippets_to_boost])
        for snippet in snippets_to_boost:
            self.current_top_snippets.insert(0, snippet)

    def add_import_trees(self, import_trees: str):
        self.import_trees += "\n" + import_trees

    def append_relevant_file_paths(self, relevant_file_paths: str):
        # do not use append, it modifies the list in place and will update it for ALL instances of RepoContextManager
        self.relevant_file_paths = self.relevant_file_paths + [relevant_file_paths]

    def set_relevant_paths(self, relevant_file_paths: list[str]):
        self.relevant_file_paths = relevant_file_paths

    def update_issue_report_and_plan(self, new_issue_report_and_plan: str):
        self.issue_report_and_plan = new_issue_report_and_plan


"""
Dump the import tree to a string
Ex:
main.py
├── database.py
│   └── models.py
└── utils.py
    └── models.py
"""

def build_full_hierarchy(
    graph: nx.DiGraph, start_node: str, k: int, prefix="", is_last=True, level=0
):
    if level > k:
        return ""
    if level == 0:
        hierarchy = f"{start_node}\n"
    else:
        hierarchy = f"{prefix}{'└── ' if is_last else '├── '}{start_node}\n"
    child_prefix = prefix + ("    " if is_last else "│   ")
    try:
        successors = {
            node
            for node, length in nx.single_source_shortest_path_length(
                graph, start_node, cutoff=1
            ).items()
            if length == 1
        }
    except Exception as e:
        print("error occured while fetching successors:", e)
        return hierarchy
    sorted_successors = sorted(successors)
    for idx, child in enumerate(sorted_successors):
        child_is_last = idx == len(sorted_successors) - 1
        hierarchy += build_full_hierarchy(
            graph, child, k, child_prefix, child_is_last, level + 1
        )
    if level == 0:
        try:
            predecessors = {
                node
                for node, length in nx.single_source_shortest_path_length(
                    graph.reverse(), start_node, cutoff=1
                ).items()
                if length == 1
            }
        except Exception as e:
            print("error occured while fetching predecessors:", e)
            return hierarchy
        sorted_predecessors = sorted(predecessors)
        for idx, parent in enumerate(sorted_predecessors):
            parent_is_last = idx == len(sorted_predecessors) - 1
            # Prepend parent hierarchy to the current node's hierarchy
            hierarchy = (
                build_full_hierarchy(graph, parent, k, "", parent_is_last, level + 1)
                + hierarchy
            )
    return hierarchy


def load_graph_from_file(filename):
    G = nx.DiGraph()
    current_node = None
    with open(filename, "r") as file:
        for line in file:
            if not line:
                continue
            if line.startswith(" "):
                line = line.strip()
                if current_node:
                    G.add_edge(current_node, line)
            else:
                line = line.strip()
                current_node = line
                if current_node:
                    G.add_node(current_node)
    return G

@file_cache(ignore_params=["rcm", "G"])
def graph_retrieval(formatted_query: str, top_k_paths: list[str], rcm: RepoContextManager, G: nx.DiGraph):
    # TODO: tune these params
    top_paths_cutoff = 25
    num_rerank = 30
    selected_paths = rcm.top_snippet_paths[:10]
    top_k_paths = top_k_paths[:top_paths_cutoff]

    snippet_scores = rcm.snippet_scores
    for snippet, score in snippet_scores.items():
        if snippet.split(":")[0] in top_k_paths:
            snippet_scores[snippet] += 1

    personalization = {}

    for snippet in selected_paths:
        personalization[snippet] = 1
    try:
        @file_cache()
        def get_distilled_file_paths(formatted_query, top_k_paths):
            personalized_pagerank_scores = nx.pagerank(G, personalization=personalization, alpha=0.85)
            unpersonalized_pagerank_scores = nx.pagerank(G, alpha=0.85)

            # tfidf style
            normalized_pagerank_scores = {path: score * log(1 / (1e-6 + unpersonalized_pagerank_scores[path])) for path, score in personalized_pagerank_scores.items()}

            top_pagerank_scores = sorted(normalized_pagerank_scores.items(), key=lambda x: x[1], reverse=True)
            
            top_pagerank_paths = [path for path, _score in top_pagerank_scores]

            distilled_file_path_list = []

            for file_path, score in top_pagerank_scores:
                if file_path.endswith(".js") and file_path.replace(".js", ".ts") in top_pagerank_paths:
                    continue
                if file_path in top_k_paths:
                    continue
                if "generated" in file_path or "mock" in file_path or "test" in file_path:
                    continue
                try:
                    rcm.cloned_repo.get_file_contents(file_path)
                except FileNotFoundError:
                    continue
                distilled_file_path_list.append(file_path)
            return distilled_file_path_list
        distilled_file_path_list = get_distilled_file_paths(formatted_query, top_k_paths)
        # Rerank once
        reranked_snippets = []
        for file_path in distilled_file_path_list[:num_rerank]:
            contents = rcm.cloned_repo.get_file_contents(file_path)
            reranked_snippets.append(Snippet(
                content=contents,
                start=0,
                end=contents.count("\n") + 1,
                file_path=file_path,
            ))
        reranked_snippets = listwise_rerank_snippets(formatted_query, reranked_snippets, prompt_type="graph")
        distilled_file_path_list[:num_rerank] = [snippet.file_path for snippet in reranked_snippets]

        return distilled_file_path_list
    except Exception as e:
        logger.error(e)
        return []

# @file_cache(ignore_params=["repo_context_manager", "override_import_graph"]) # can't cache this because rcm is stateful
def integrate_graph_retrieval(formatted_query: str, repo_context_manager: RepoContextManager, override_import_graph: nx.DiGraph = None):
    num_graph_retrievals = 25
    repo_context_manager, import_graph = parse_query_for_files(formatted_query, repo_context_manager)
    if override_import_graph:
        import_graph = override_import_graph
    if import_graph:
        # Graph retrieval can fail and return [] if the graph is not found or pagerank does not converge
        # Happens especially when graph has multiple components
        graph_retrieved_files = graph_retrieval(formatted_query, sorted(repo_context_manager.top_snippet_paths), repo_context_manager, import_graph) # sort input for caching
        if graph_retrieved_files:
            sorted_snippets = sorted(
                repo_context_manager.snippets,
                key=lambda snippet: repo_context_manager.snippet_scores[snippet.denotation],
                reverse=True,
            )
            snippets = []
            for file_path in graph_retrieved_files:
                for snippet in sorted_snippets[50 - num_graph_retrievals:]:
                    if snippet.file_path == file_path:
                        snippets.append(snippet)
                        break
            graph_retrieved_files = graph_retrieved_files[:num_graph_retrievals]
            repo_context_manager.read_only_snippets = snippets[:len(graph_retrieved_files)]
            repo_context_manager.current_top_snippets = repo_context_manager.current_top_snippets[:50 - num_graph_retrievals]
    return repo_context_manager, import_graph

# add import trees for any relevant_file_paths (code files that appear in query)
def build_import_trees(
    rcm: RepoContextManager,
    import_graph: nx.DiGraph,
    override_import_graph: nx.DiGraph = None,
) -> tuple[RepoContextManager]:
    if import_graph is None and override_import_graph is None:
        return rcm
    if override_import_graph:
        import_graph = override_import_graph
    # if we have found relevant_file_paths in the query, we build their import trees
    code_files_in_query = rcm.relevant_file_paths
    # graph_retrieved_files = graph_retrieval(rcm.top_snippet_paths, rcm, import_graph)[:15]
    graph_retrieved_files = [snippet.file_path for snippet in rcm.read_only_snippets]
    if code_files_in_query:
        for file in code_files_in_query:
            # fetch direct parent and children
            representation = (
                f"\nThe file '{file}' has the following import structure: \n"
                + build_full_hierarchy(import_graph, file, 2)
            )
            if graph_retrieved_files:
                representation += "\n\nThe following modules may contain helpful services or utility functions:\n- " + "\n- ".join(graph_retrieved_files)
            rcm.add_import_trees(representation)
    # if there are no code_files_in_query, we build import trees for the top 5 snippets
    else:
        for snippet in rcm.current_top_snippets[:5]:
            file_path = snippet.file_path
            representation = (
                f"\nThe file '{file_path}' has the following import structure: \n"
                + build_full_hierarchy(import_graph, file_path, 2)
            )
            if graph_retrieved_files:
                representation += "\n\nThe following modules may contain helpful services or utility functions:\n- " + "\n-".join(graph_retrieved_files)
            rcm.add_import_trees(representation)
    return rcm


# add any code files that appear in the query to current_top_snippets
def add_relevant_files_to_top_snippets(rcm: RepoContextManager) -> RepoContextManager:
    code_files_in_query = rcm.relevant_file_paths
    for file in code_files_in_query:
        current_top_snippet_paths = [
            snippet.file_path for snippet in rcm.current_top_snippets
        ]
        # if our mentioned code file isnt already in the current_top_snippets we add it
        if file not in current_top_snippet_paths:
            try:
                code_snippets = [
                    snippet for snippet in rcm.snippets if snippet.file_path == file
                ]
                rcm.boost_snippets_to_top(code_snippets)
            except Exception as e:
                logger.error(
                    f"Tried to add code file found in query but recieved error: {e}, skipping and continuing to next one."
                )
    return rcm


# fetch all files mentioned in the user query
def parse_query_for_files(
    query: str, rcm: RepoContextManager
) -> tuple[RepoContextManager, nx.DiGraph]:
    # use cloned_repo to attempt to find any files names that appear in the query
    repo_full_name = rcm.cloned_repo.repo_full_name
    repo_name = repo_full_name.split("/")[-1]
    repo_group_name = repo_full_name.split("/")[0]
    code_files_to_add = set([])
    code_files_to_check = set(list(rcm.cloned_repo.get_file_list()))
    code_files_uri_encoded = [
        urllib.parse.quote(file_path) for file_path in code_files_to_check
    ]
    # check if any code files are mentioned in the query
    for file, file_uri_encoded in zip(code_files_to_check, code_files_uri_encoded):
        if file in query or file_uri_encoded in query:
            code_files_to_add.add(file)
    for code_file in code_files_to_add:
        rcm.append_relevant_file_paths(code_file)
    # only for enterprise
    try:
        pathing = (
            f"{repo_group_name}_import_graphs/{repo_name}/{repo_name}_import_tree.txt"
        )
        if not os.path.exists(pathing):
            return rcm, None
        graph = load_graph_from_file(pathing)
    except Exception as e:
        logger.error(
            f"Error loading import tree: {e}, skipping step and setting import_tree to empty string"
        )
        return rcm, None
    files = set(list(graph.nodes()))
    files_uri_encoded = [urllib.parse.quote(file_path) for file_path in files]
    for file, file_uri_encoded in zip(files, files_uri_encoded):
        if (file in query or file_uri_encoded in query) and (
            file not in code_files_to_add
        ):
            rcm.append_relevant_file_paths(file)
    return rcm, graph


# do not ignore repo_context_manager
# @file_cache(ignore_params=["seed", "ticket_progress", "chat_logger"])
def get_relevant_context(
    query: str,
    repo_context_manager: RepoContextManager,
    seed: int = None,
    import_graph: nx.DiGraph = None,
    num_rollouts: int = NUM_ROLLOUTS,
    ticket_progress = None,
    chat_logger = None,
) -> RepoContextManager:
    logger.info("Seed: " + str(seed))
    try:
        # for any code file mentioned in the query, build its import tree - This is currently not used
        repo_context_manager = build_import_trees(
            repo_context_manager,
            import_graph,
        )
        # for any code file mentioned in the query add it to the top relevant snippets
        repo_context_manager = add_relevant_files_to_top_snippets(repo_context_manager)
        # add relevant files to dir_obj inside repo_context_manager, this is in case dir_obj is too large when as a string
        repo_context_manager.dir_obj.add_relevant_files(
            repo_context_manager.relevant_file_paths
        )

        user_prompt = repo_context_manager.format_context(
            unformatted_user_prompt=unformatted_user_prompt,
            query=query,
        )
        return repo_context_manager # Temporarily disabled context
        chat_gpt = ChatGPT()
        chat_gpt.messages = [Message(role="system", content=sys_prompt)]
        old_relevant_snippets = deepcopy(repo_context_manager.current_top_snippets)
        old_read_only_snippets = deepcopy(repo_context_manager.read_only_snippets)
        try:
            repo_context_manager = context_dfs(
                user_prompt,
                repo_context_manager,
                problem_statement=query,
                num_rollouts=num_rollouts,
            )
        except openai.BadRequestError as e:  # sometimes means that run has expired
            logger.exception(e)
        repo_context_manager.current_top_snippets.extend(old_relevant_snippets)
        repo_context_manager.read_only_snippets.extend(old_read_only_snippets)
        return repo_context_manager
    except Exception as e:
        logger.exception(e)
        return repo_context_manager


def update_assistant_conversation(
    run: Run,
    thread: Thread,
    ticket_progress: TicketProgress,
    repo_context_manager: RepoContextManager,
):
    assistant_conversation = AssistantConversation.from_ids(
        assistant_id=run.assistant_id,
        run_id=run.id,
        thread_id=thread.id,
    )
    if ticket_progress:
        if assistant_conversation:
            ticket_progress.search_progress.pruning_conversation = (
                assistant_conversation
            )
        ticket_progress.search_progress.repo_tree = str(repo_context_manager.dir_obj)
        ticket_progress.search_progress.final_snippets = (
            repo_context_manager.current_top_snippets
        )
        ticket_progress.save()


CLAUDE_MODEL = "claude-3-haiku-20240307"


def validate_and_parse_function_calls(
    function_calls_string: str, chat_gpt: ChatGPT
) -> list[AnthropicFunctionCall]:
    function_calls = AnthropicFunctionCall.mock_function_calls_from_string(
        function_calls_string.strip("\n") + "\n</function_call>"
    )  # add end tag
    if len(function_calls) > 0:
        chat_gpt.messages[-1].content = (
            chat_gpt.messages[-1].content.rstrip("\n") + "\n</function_call>"
        )  # add end tag to assistant message
        return function_calls

    # try adding </invoke> tag as well
    function_calls = AnthropicFunctionCall.mock_function_calls_from_string(
        function_calls_string.strip("\n") + "\n</invoke>\n</function_call>"
    )
    if len(function_calls) > 0:
        # update state of chat_gpt
        chat_gpt.messages[-1].content = (
            chat_gpt.messages[-1].content.rstrip("\n") + "\n</invoke>\n</function_call>"
        )
        return function_calls
    # try adding </parameters> tag as well
    function_calls = AnthropicFunctionCall.mock_function_calls_from_string(
        function_calls_string.strip("\n")
        + "\n</parameters>\n</invoke>\n</function_call>"
    )
    if len(function_calls) > 0:
        # update state of chat_gpt
        chat_gpt.messages[-1].content = (
            chat_gpt.messages[-1].content.rstrip("\n")
            + "\n</parameters>\n</invoke>\n</function_call>"
        )
    return function_calls


def handle_function_call(
    repo_context_manager: RepoContextManager, function_call: AnthropicFunctionCall, llm_state: dict[str, str]
):
    function_name = function_call.function_name
    function_input = function_call.function_parameters
    logger.info(f"Tool Call: {function_name} {function_input}")
    file_path = function_input.get("file_path", None)
    valid_path = False
    output_prefix = f"Output for {function_name}:\n"
    output = ""
    current_top_snippets_string = "\n".join(
        list(dict.fromkeys([snippet.file_path for snippet in repo_context_manager.current_top_snippets]))
    )
    if function_name == "code_search":
        code_entity = f'"{function_input["code_entity"]}"'  # handles cases with two words
        code_entity = escape_ripgrep(code_entity) # escape special characters
        rg_command = [
            "rg",
            "-n",
            "-i",
            code_entity,
            repo_context_manager.cloned_repo.repo_dir,
        ]
        try:
            result = subprocess.run(
                " ".join(rg_command), text=True, shell=True, capture_output=True
            )
            rg_output = result.stdout
            if rg_output:
                # post process rip grep output to be more condensed
                rg_output_pretty, file_output_dict, file_to_num_occurrences = post_process_rg_output(
                    repo_context_manager.cloned_repo.repo_dir, SweepConfig(), rg_output
                )
                # return results first by occurrences then by alphabetical order
                non_stored_files = sorted([
                    file_path
                    for file_path in file_output_dict
                    if file_path not in repo_context_manager.top_snippet_paths
                ], key=lambda x: (-file_to_num_occurrences[x], x))
                non_stored_files = [file_path + f" ({file_to_num_occurrences[file_path]} occurrences)" for file_path in non_stored_files]
                non_stored_files_string = "These search results have not been stored:\n<non_stored_search_results>\n" + "\n".join(non_stored_files) + "\n</non_stored_search_results>\n" if non_stored_files else "All of the files above have already been stored. Search for a new term.\n"
                if len(file_output_dict) <= 10:
                    output = (
                        f"SUCCESS: Here are the code_search results:\n<code_search_results>\n{rg_output_pretty}<code_search_results>\n" +
                        non_stored_files_string + 
                        "Use the `view_files` tool to read the most relevant non-stored files. Use `store_file` to add any important non-stored files to the context. DO NOT VIEW FILES THAT HAVE BEEN STORED."
                    )
                else:
                    output = (
                        f"SUCCESS: Here are the code_search results:\n<code_search_results>\n{rg_output_pretty}<code_search_results>\n" +
                        non_stored_files_string + "Prioritize viewing the non-stored files with the most occurrences. Use the `view_files` tool to read the most relevant non-stored files. Use `store_file` to add any important non-stored files to the context. DO NOT VIEW FILES THAT HAVE BEEN STORED."
                    )
                # too many prompt it to search more specific
            else:
                output = f"FAILURE: No results found for code_entity: {code_entity} in the entire codebase. Please try a new code_entity. Consider trying different whitespace or a truncated version of this code_entity."
        except Exception as e:
            logger.error(
                f"FAILURE: An Error occured while trying to find the code_entity {code_entity}: {e}"
            )
            output = f"FAILURE: No results found for code_entity: {code_entity} in the entire codebase. Please try a new code_entity. Consider trying different whitespace or a truncated version of this code_entity."
    elif function_name == "view_files":
        output = ""
        all_viewed_files = [function_input.get("first_file_path", ""), function_input.get("second_file_path", ""), function_input.get("file_path", "")]
        all_viewed_files = [file_path for file_path in all_viewed_files if file_path]
        for file_path in all_viewed_files:
            try:
                file_contents = repo_context_manager.cloned_repo.get_file_contents(
                    file_path
                )
                # check if file has been viewed already
                # function_call_history = llm_state.get("function_call_history", [])
                # # unnest 2d list
                # previous_function_calls = [
                #     call for sublist in function_call_history for call in sublist
                # ]
                # previously_viewed_files = list(dict.fromkeys(previously_viewed_files))
                # if file_path in previously_viewed_files:
                #     previously_viewed_files_str = "\n".join(previously_viewed_files)
                #     output = f"WARNING: `{file_path}` has already been viewed. Please refer to the file in your previous function call. These files have already been viewed:\n{previously_viewed_files_str}"
                if file_path not in [snippet.file_path for snippet in repo_context_manager.current_top_snippets]:
                    output += f'SUCCESS: Here are the contents of `{file_path}`:\n<source>\n{file_contents}\n</source>\nYou can use the `store_file` tool to add this file to the context.'
                else:
                    output += f"FAILURE: {file_path} has already been stored. Please view a new file."
            except FileNotFoundError:
                file_contents = ""
                similar_file_paths = "\n".join(
                    [
                        f"- {path}"
                        for path in repo_context_manager.cloned_repo.get_similar_file_paths(
                            file_path
                        )
                    ]
                )
                output += f"FAILURE: {file_path} does not exist. Did you mean:\n{similar_file_paths}\n"
    elif function_name == "store_file":
        try:
            file_contents = repo_context_manager.cloned_repo.get_file_contents(
                file_path
            )
            valid_path = True
        except Exception:
            file_contents = ""
            similar_file_paths = "\n".join(
                [
                    f"- {path}"
                    for path in repo_context_manager.cloned_repo.get_similar_file_paths(
                        file_path
                    )
                ]
            )
            output = f"FAILURE: This file path does not exist. Did you mean:\n{similar_file_paths}"
        else:
            snippet = Snippet(
                file_path=file_path,
                start=0,
                end=len(file_contents.splitlines()),
                content=file_contents,
            )
            if snippet.file_path in current_top_snippets_string:
                output = f"FAILURE: {get_stored_files(repo_context_manager)}"
            else:
                repo_context_manager.add_snippets([snippet])
                current_top_snippets_string = "\n".join(
                    list(dict.fromkeys([snippet.file_path for snippet in repo_context_manager.current_top_snippets]))
                )
                output = (
                    f"SUCCESS: {file_path} was added to the stored_files. It will be used as a reference or modified to resolve the issue."
                    if valid_path
                    else f"FAILURE: The file path '{file_path}' does not exist. Please check the path and try again."
                )
    elif function_name == "submit":
        plan = function_input.get("plan")
        repo_context_manager.update_issue_report_and_plan(f"# Highly Suggested Plan:\n\n{plan}\n\n")
        output = PLAN_SUBMITTED_MESSAGE
    else:
        output = f"FAILURE: Invalid tool name {function_name}"
    analysis = (
        function_input["analysis"] if "analysis" in function_input else ""
    )
    logger.info(
        f"Tool Call: {function_name}\n{analysis}\n{output}"
    )
    return (output_prefix + output)


reflections_prompt_prefix = """
CRITICAL FEEDBACK - READ CAREFULLY AND ADDRESS ALL POINTS
<critical_feedback_to_address>
Here is the feedback from your previous attempt. You MUST read this extremely carefully and follow ALL of the reviewer's advice. If they tell you to store specific files, view store them first. If you do not fully address this feedback you will fail to retrieve all of the relevant files.
{all_reflections}
</critical_feedback_to_address>"""

reflection_prompt = """<attempt_and_feedback_{idx}>
<previous_files_stored>
Files stored from previous attempt:
{files_read}
</previous_files_stored>
<rating>
Rating from previous attempt: {score} / 10
</rating>
<feedback>
Reviewer feedback on previous attempt:
{reflections_string}
</feedback>
</attempt_and_feedback_{idx}>"""

def format_reflections(reflections_to_gathered_files: dict[str, tuple[list[str], int]]) -> str:
    formatted_reflections_prompt = ""
    if not reflections_to_gathered_files:
        return formatted_reflections_prompt
    all_reflections_string = "\n"
    # take only the MAX_REFLECTIONS sorted by score
    top_reflections = sorted(
        reflections_to_gathered_files.items(), key=lambda x: x[1][1] * 100 + len(x[1][0]), reverse=True # break ties by number of files stored
    )[:MAX_REFLECTIONS]
    for idx, (reflection, (gathered_files, score)) in enumerate(top_reflections):
        formatted_reflection = reflection_prompt.format(
            files_read="\n".join(gathered_files),
            reflections_string=reflection,
            score=str(score),
            idx=str(idx + 1),
        )
        all_reflections_string += f"\n{formatted_reflection}"
    formatted_reflections_prompt = reflections_prompt_prefix.format(
        all_reflections=all_reflections_string
    )
    return formatted_reflections_prompt

def render_all_attempts(function_call_histories: list[list[list[AnthropicFunctionCall]]]) -> str:
    formatted_attempts = ""
    for idx, function_call_history in enumerate(function_call_histories):
        formatted_function_calls = render_function_calls_for_attempt(function_call_history)
        formatted_attempts += f"<attempt_{idx}>\n{formatted_function_calls}\n</attempt_{idx}>"
    return formatted_attempts

def render_function_calls_for_attempt(function_call_history: list[list[AnthropicFunctionCall]]) -> str:
    formatted_function_calls = ""
    idx = 0
    for function_calls in function_call_history:
        for function_call in function_calls:
            function_call.function_parameters.pop("analysis", None) # remove analysis
            function_call_cleaned_string = function_call.function_name + " | " + "\n".join([str(k) + " | " + str(v) for k, v in function_call.function_parameters.items()])
            formatted_function_calls += f"- {function_call_cleaned_string}\n"
        if function_calls:
            idx += 1
    return formatted_function_calls

def get_stored_files(repo_context_manager: RepoContextManager) -> str:
    fetched_files_that_are_stored = list(dict.fromkeys([snippet.file_path for snippet in repo_context_manager.current_top_snippets]))
    joined_files_string = "\n".join(fetched_files_that_are_stored)
    stored_files_string = f'The following files have been stored already. DO NOT CALL THE STORE OR VIEW TOOLS ON THEM AGAIN. \n<stored_files>\n{joined_files_string}\n</stored_files>\n' if fetched_files_that_are_stored else ""
    return stored_files_string

def search_for_context_with_reflection(repo_context_manager: RepoContextManager, reflections_to_read_files: dict[str, tuple[list[str], int]], user_prompt: str, rollout_function_call_histories: list[list[list[AnthropicFunctionCall]]], problem_statement: str) -> tuple[list[Message], list[list[AnthropicFunctionCall]]]:
    try:
        _, function_call_history = perform_rollout(repo_context_manager, reflections_to_read_files, user_prompt)
        rollout_function_call_histories.append(function_call_history)
    except Exception as e:
        logger.error(f"Error in perform_rollout: {e}")
    rollout_stored_files = [snippet.file_path for snippet in repo_context_manager.current_top_snippets]
    # truncated_message_results = message_results[1:] # skip system prompt
    # joined_messages = "\n\n".join([message.content for message in truncated_message_results])
    # overall_score, message_to_contractor = EvaluatorAgent().evaluate_run(
    #     problem_statement=problem_statement, 
    #     run_text=joined_messages,
    #     stored_files=rollout_stored_files,
    # )
    return 0, "", repo_context_manager, rollout_stored_files

def perform_rollout(repo_context_manager: RepoContextManager, reflections_to_gathered_files: dict[str, tuple[list[str], int]], user_prompt: str) -> list[Message]:
    function_call_history = []
    formatted_reflections_prompt = format_reflections(reflections_to_gathered_files)
    updated_user_prompt = user_prompt + formatted_reflections_prompt
    chat_gpt = ChatGPT()
    chat_gpt.messages = [Message(role="system", content=sys_prompt + formatted_reflections_prompt)]
    function_calls_string = chat_gpt.chat_anthropic(
        content=updated_user_prompt,
        stop_sequences=["</function_call>"],
        model=CLAUDE_MODEL,
        message_key="user_request",
        assistant_message_content="<function_call>",
    )
    bad_call_count = 0
    llm_state = {} # persisted across one rollout
    llm_state["function_call_history"] = {}
    for _ in range(MAX_ITERATIONS):
        function_calls = validate_and_parse_function_calls(
            function_calls_string, chat_gpt
        )
        function_outputs = ""
        for function_call in function_calls[:MAX_PARALLEL_FUNCTION_CALLS]:
            function_outputs += handle_function_call(repo_context_manager, function_call, llm_state) + "\n"
            logger.info(f"Function outputs: {function_outputs}")
            logger.info("Function call: " + str(function_call))
            llm_state["function_call_history"] = function_call_history
            if PLAN_SUBMITTED_MESSAGE in function_outputs:
                return chat_gpt.messages, function_call_history
        function_call_history.append(function_calls)
        if len(function_calls) == 0:
            function_outputs = "REMINDER: No function calls were made or your last function call was incorrectly formatted. The correct syntax for function calling is this:\n" \
                + "<function_call>\n<invoke>\n<tool_name>tool_name</tool_name>\n<parameters>\n<param_name>param_value</param_name>\n</parameters>\n</invoke>\n</function_call>" + "\nRemember to gather ALL relevant files. " + get_stored_files(repo_context_manager)
            bad_call_count += 1
        if function_outputs.startswith("FAILURE"):
            bad_call_count += 1
        if bad_call_count >= NUM_BAD_FUNCTION_CALLS:
            return chat_gpt.messages, function_call_history
        if len(function_calls) > MAX_PARALLEL_FUNCTION_CALLS:
            remaining_function_calls = function_calls[MAX_PARALLEL_FUNCTION_CALLS:]
            remaining_function_calls_string = mock_function_calls_to_string(remaining_function_calls)
            function_outputs += "WARNING: You requested more than 1 function call at once. Only the first function call has been processed. The unprocessed function calls were:\n<unprocessed_function_call>\n" + remaining_function_calls_string + "\n</unprocessed_function_call>"
        try:
            function_calls_string = chat_gpt.chat_anthropic(
                content=function_outputs,
                model=CLAUDE_MODEL,
                stop_sequences=["</function_call>"],
                assistant_message_content="<function_call>",
            )
        except Exception as e:
            logger.error(f"Error in chat_anthropic: {e}")
            # return all but the last message because it likely causes an error
            return chat_gpt.messages[:-1], function_call_history
    return chat_gpt.messages, function_call_history

def context_dfs(
    user_prompt: str,
    repo_context_manager: RepoContextManager,
    problem_statement: str,
    num_rollouts: int,
) -> bool | None:
    # initial function call
    reflections_to_read_files = {}
    rollouts_to_scores_and_rcms = {}
    rollout_function_call_histories = []
    for rollout_idx in range(num_rollouts):
        overall_score, message_to_contractor, repo_context_manager, rollout_stored_files = search_for_context_with_reflection(
            repo_context_manager=repo_context_manager,
            reflections_to_read_files=reflections_to_read_files,
            user_prompt=user_prompt,
            rollout_function_call_histories=rollout_function_call_histories,
            problem_statement=problem_statement
        )
        logger.info(f"Completed run {rollout_idx} with score: {overall_score} and reflection: {message_to_contractor}")
        if overall_score is None or message_to_contractor is None:
            continue # can't get any reflections here
        # reflections_to_read_files[message_to_contractor] = rollout_stored_files, overall_score
        rollouts_to_scores_and_rcms[rollout_idx] = (overall_score, repo_context_manager)
        if overall_score >= SCORE_THRESHOLD and len(rollout_stored_files) > STOP_AFTER_SCORE_THRESHOLD_IDX:
            break
    # if we reach here, we have not found a good enough solution
    # select rcm from the best rollout
    logger.info(f"{render_all_attempts(rollout_function_call_histories)}")
    all_scores_and_rcms = list(rollouts_to_scores_and_rcms.values())
    best_score, best_rcm = max(all_scores_and_rcms, key=lambda x: x[0] * 100 + len(x[1].current_top_snippets)) # sort first on the highest score, break ties with length of current_top_snippets
    for score, rcm in all_scores_and_rcms:
        logger.info(f"Rollout score: {score}, Rollout files: {[snippet.file_path for snippet in rcm.current_top_snippets]}")
    logger.info(f"Best score: {best_score}, Best files: {[snippet.file_path for snippet in best_rcm.current_top_snippets]}")
    return best_rcm

if __name__ == "__main__":
    try:
        from sweepai.utils.github_utils import get_installation_id
        from sweepai.utils.ticket_utils import prep_snippets

        organization_name = "sweepai"
        installation_id = get_installation_id(organization_name)
        cloned_repo = ClonedRepo("sweepai/sweep", installation_id, "main")
        query = "allow 'sweep.yaml' to be read from the user/organization's .github repository. this is found in client.py and we need to change this to optionally read from .github/sweep.yaml if it exists there"
        # golden response is
        # sweepai/handlers/create_pr.py:401-428
        # sweepai/config/client.py:178-282
        ticket_progress = TicketProgress(
            tracking_id="test",
        )
        repo_context_manager = prep_snippets(cloned_repo, query, ticket_progress)
        rcm = get_relevant_context(
            query,
            repo_context_manager,
            ticket_progress,
            chat_logger=ChatLogger({"username": "wwzeng1"}),
        )
        for snippet in rcm.current_top_snippets:
            print(snippet.denotation)
    except Exception as e:
        logger.error(f"context_pruning.py failed to run successfully with error: {e}")
        raise e



from loguru import logger

from sweepai.config.client import (
    RESET_FILE,
    REVERT_CHANGED_FILES_TITLE,
    SWEEP_BAD_FEEDBACK,
    SWEEP_GOOD_FEEDBACK,
    SweepConfig,
    get_documentation_dict,
)
from sweepai.config.server import DISCORD_FEEDBACK_WEBHOOK_URL
from sweepai.core.context_pruning import get_relevant_context
from sweepai.core.entities import NoFilesException, SandboxResponse
from sweepai.core.sweep_bot import SweepBot, get_files_to_change, validate_file_change_requests

# from sandbox.sandbox_utils import Sandbox
from sweepai.handlers.create_pr import GITHUB_LABEL_NAME, create_pr_changes
from sweepai.utils.buttons import Button, ButtonList, create_action_buttons
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.github_utils import ClonedRepo, get_github_client
from sweepai.utils.prompt_constructor import HumanMessagePrompt
from sweepai.utils.ticket_utils import prep_snippets


def make_pr(
    title,
    repo_description,
    summary,
    repo_full_name,
    installation_id,
    user_token,
    use_faster_model,
    username,
    chat_logger: ChatLogger,
    branch_name=None,
    rule=None,
):
    chat_logger.data["title"] = title
    _, repo_name = repo_full_name.split("/")
    # heavily copied code from on_ticket
    cloned_repo = ClonedRepo(
        repo_full_name,
        installation_id=installation_id,
        token=user_token,
        branch=branch_name,
    )
    logger.info("Fetching relevant files...")
    search_query = (title + summary).strip("\n")
    formatted_query = (f"{title.strip()}\n{summary.strip()}").strip("\n")
    repo_context_manager = prep_snippets(cloned_repo, search_query)
    repo_context_manager = get_relevant_context(
        formatted_query,
        repo_context_manager,
        chat_logger=chat_logger,
    )
    snippets = repo_context_manager.current_top_snippets
    tree = str(repo_context_manager.dir_obj)
    message_summary = summary
    # removed external search as it provides no real value and only adds noise
    # external_results = ExternalSearcher.extract_summaries(message_summary)
    # if external_results:
    #     message_summary += "\n\n" + external_results
    get_documentation_dict(cloned_repo.repo)
    human_message = HumanMessagePrompt(
        repo_name=repo_name,
        username=username,
        repo_description=repo_description.strip(),
        title=title,
        summary=message_summary,
        snippets=snippets,
        tree=tree,
    )

    _user_token, g = get_github_client(installation_id)
    repo = g.get_repo(repo_full_name)
    sweep_bot = SweepBot.from_system_message_content(
        human_message=human_message,
        repo=repo,
        is_reply=False,
        chat_logger=chat_logger,
        cloned_repo=cloned_repo,
    )
    file_change_requests, plan = get_files_to_change(
        relevant_snippets=repo_context_manager.current_top_snippets,
        read_only_snippets=repo_context_manager.read_only_snippets,
        problem_statement=formatted_query,
        repo_name=repo_name,
    )
    validate_file_change_requests(file_change_requests, repo_context_manager.cloned_repo)
    file_change_requests = sweep_bot.validate_file_change_requests(
        file_change_requests, branch_name
    )
    pull_request = sweep_bot.generate_pull_request()
    generator = create_pr_changes(
        file_change_requests,
        pull_request,
        sweep_bot,
        username,
        installation_id,
        chat_logger=chat_logger,
        base_branch=branch_name,
    )
    response = {"error": NoFilesException()}
    changed_files = []
    for item in generator:
        if isinstance(item, dict):
            response = item
            break
        (
            file_change_request,
            changed_file,
            sandbox_response,
            commit,
            file_change_requests,
        ) = item
        if changed_file:
            changed_files.append(file_change_request.filename)
        sandbox_response: SandboxResponse | None = sandbox_response
    pr_changes = response["pull_request"]
    pr_actions_message = (
        create_action_buttons(
            [
                SWEEP_GOOD_FEEDBACK,
                SWEEP_BAD_FEEDBACK,
            ],
            header="### PR Feedback (click)\n",
        )
        + "\n"
        if DISCORD_FEEDBACK_WEBHOOK_URL is not None
        else ""
    )
    rule_description = (
        f'### I created this PR to address this rule: \n"{rule}"\n'
        if rule
        else "I created this PR to fix the failing GitHub Actions."
    )
    pr = repo.create_pull(
        title=pr_changes.title,
        body=pr_actions_message + rule_description + pr_changes.body,
        head=pr_changes.pr_head,
        base=branch_name if branch_name else SweepConfig.get_branch(repo),
    )
    pr.add_to_assignees(username)
    buttons = []
    for changed_file in changed_files:
        buttons.append(Button(label=f"{RESET_FILE} {changed_file}"))
    revert_buttons_list = ButtonList(buttons=buttons, title=REVERT_CHANGED_FILES_TITLE)
    pr.create_issue_comment(revert_buttons_list.serialize())
    pr.add_to_labels(GITHUB_LABEL_NAME)

    return pr



import os
import re


def check_comments_presence(file_path: str, new_code: str) -> bool:
    _, file_extension = os.path.splitext(file_path)
    comment_patterns = {
        ".py": "#",
        ".js": "//",
        ".ts": "//",
        ".jsx": "//",
        ".tsx": "//",
        ".java": "//",
        ".c": "/*",
        ".cpp": "//",
        ".cs": "//",
        ".php": "//",
        ".swift": "//",
        ".rb": "#",
    }
    if file_extension not in comment_patterns:
        return False
    comment_pattern = comment_patterns.get(file_extension, "")
    return bool(re.search(comment_pattern, new_code))



from __future__ import annotations
from dataclasses import dataclass
import re


def convert_openai_function_to_anthropic_prompt(function: dict) -> str:
    unformatted_prompt = """<tool_description>
<tool_name>{tool_name}</tool_name>
<description>
{description}
</description>
<parameters>
{parameters}
</parameters>
</tool_description>"""
    unformatted_parameter = """<parameter>
<name>{parameter_name}</name>
<type>{parameter_type}</type>
<description>{parameter_description}</description>
</parameter>"""
    parameters_strings = []
    
    for parameter_name, parameter_dict in function["parameters"]["properties"].items():
        parameters_strings.append(unformatted_parameter.format(
            parameter_name=parameter_name,
            parameter_type=parameter_dict["type"],
            parameter_description=parameter_dict["description"],
        ))
    return unformatted_prompt.format(
        tool_name=function["name"],
        description=function["description"],
        parameters="\n".join(parameters_strings),
    )

def convert_all_functions(functions: list) -> str:
    # convert all openai functions to print anthropic prompt
    for function in functions:
        print(convert_openai_function_to_anthropic_prompt(function))

@dataclass
class AnthropicFunctionCall:
    function_name: str
    function_parameters: dict[str, str]

    def to_string(self) -> str:
        function_call_string = "<invoke>\n"
        function_call_string += f"<tool_name>{self.function_name}</tool_name>\n"
        function_call_string += "<parameters>\n"
        for param_name, param_value in self.function_parameters.items():
            function_call_string += f"<{param_name}>\n{param_value}\n</{param_name}>\n"
        function_call_string += "</parameters>\n"
        function_call_string += "</invoke>"
        return function_call_string

    @staticmethod
    def mock_function_calls_from_string(function_calls_string: str) -> list[AnthropicFunctionCall]:
        function_calls = []

        # Regular expression patterns
        function_name_pattern = r'<tool_name>(.*?)</tool_name>'
        parameters_pattern = r'<parameters>(.*?)</parameters>'
        parameter_pattern = r'<(.*?)>(.*?)<\/\1>'
        
        # Extract function calls
        function_call_matches = re.findall(r'<invoke>(.*?)</invoke>', function_calls_string, re.DOTALL)
        for function_call_match in function_call_matches:
            # Extract function name
            function_name_match = re.search(function_name_pattern, function_call_match)
            function_name = function_name_match.group(1) if function_name_match else None

            # Extract parameters section
            parameters_match = re.search(parameters_pattern, function_call_match, re.DOTALL)
            parameters_section = parameters_match.group(1) if parameters_match else ''

            # Extract parameters within the parameters section
            parameter_matches = re.findall(parameter_pattern, parameters_section, re.DOTALL)
            function_parameters = {}
            for param in parameter_matches:
                parameter_name = param[0]
                parameter_value = param[1]
                function_parameters[parameter_name] = parameter_value.strip()

            if function_name and function_parameters != {}:
                function_calls.append(AnthropicFunctionCall(function_name, function_parameters))

        return function_calls

def mock_function_calls_to_string(function_calls: list[AnthropicFunctionCall]) -> str:
    function_calls_string = "<function_call>\n"
    for function_call in function_calls:
        function_calls_string += function_call.to_string() + "\n"
    function_calls_string += "</function_call>"
    return function_calls_string

if __name__ == "__main__":    
    test_str = """<function_call>
<invoke>
<tool_name>submit_report_and_plan</tool_name>
<parameters>
<report>
The main API implementation for the Sweep application is in the `sweepai/api.py` file. This file handles various GitHub events, such as pull requests, issues, and comments, and triggers corresponding actions.

The `PRChangeRequest` class, defined in the `sweepai/core/entities.py` file, is used to encapsulate information about a pull request change, such as the comment, repository, and user information. This class is utilized throughout the `sweepai/api.py` file to process and respond to the different GitHub events.

To solve the user request, the following plan should be followed:

1. Carefully review the `sweepai/api.py` file to understand how the different GitHub events are handled and the corresponding actions that are triggered.
2. Analyze the usage of the `PRChangeRequest` class in the `sweepai/api.py` file to understand how it is used to process pull request changes.
3. Determine the specific issue or feature that needs to be implemented or fixed based on the user request.
4. Implement the necessary changes in the `sweepai/api.py` file, utilizing the `PRChangeRequest` class as needed.
5. Ensure that the changes are thoroughly tested and that all relevant cases are covered.
6. Submit the changes for review and deployment.
</report>
<plan>
1. Review the `sweepai/api.py` file to understand the overall structure and flow of the application, focusing on how GitHub events are handled and the corresponding actions that are triggered.
2. Analyze the usage of the `PRChangeRequest` class in the `sweepai/api.py` file to understand how it is used to process pull request changes, including the information it encapsulates and the various methods that operate on it.
3. Determine the specific issue or feature that needs to be implemented or fixed based on the user request. This may involve identifying the relevant GitHub event handlers and the corresponding logic that needs to be modified.
4. Implement the necessary changes in the `sweepai/api.py` file, utilizing the `PRChangeRequest` class as needed to process the pull request changes. This may include adding new event handlers, modifying existing ones, or enhancing the functionality of the `PRChangeRequest` class.
5. Thoroughly test the changes to ensure that all relevant cases are covered, including edge cases and error handling. This may involve writing additional unit tests or integration tests to validate the functionality.
6. Once the changes have been implemented and tested, submit the modified `sweepai/api.py` file for review and deployment.
</plan>
</parameters>
</invoke>
</function_call>"""

    function_calls = AnthropicFunctionCall.mock_function_calls_from_string(test_str)
    for function_call in function_calls:
        print(function_call)
        print(function_call.to_string())
    print(mock_function_calls_to_string(function_calls))


def read_file_with_fallback_encodings(
    file_path, encodings=["utf-8", "windows-1252", "iso-8859-1"]
):
    for encoding in encodings:
        try:
            with open(file_path, "r", encoding=encoding) as file:
                return file.read()
        except UnicodeDecodeError:
            continue
    raise UnicodeDecodeError(
        f"Could not decode {file_path} with any of the specified encodings: {encodings}"
    )


import unittest
import unittest.mock

from sweepai.utils.buttons import (
    Button,
    ButtonList,
    check_button_activated,
    check_button_title_match,
    create_action_buttons,
    create_button,
    get_toggled_state,
)


class TestButtons(unittest.TestCase):
    def test_create_button(self):
        self.assertEqual(create_button("Test", False), "- [ ] Test")
        self.assertEqual(create_button("Test", True), "- [x] Test")

    def test_create_action_buttons(self):
        labels = ["Test1", "Test2"]
        expected_output = "## Actions (click)\n- [ ] Test1\n- [ ] Test2"
        self.assertEqual(create_action_buttons(labels), expected_output)

    def test_get_toggled_state(self):
        changes_request = unittest.mock.Mock()
        changes_request.body_from = "- [x] Test"
        self.assertTrue(get_toggled_state("Test", changes_request))

    def test_check_button_activated(self):
        changes_request = unittest.mock.Mock()
        changes_request.body_from = "- [ ] Test"
        self.assertTrue(check_button_activated("Test", "- [x] Test", changes_request))

    def test_check_button_title_match(self):
        changes_request = unittest.mock.Mock()
        changes_request.body_from = "Test"
        self.assertTrue(check_button_title_match("Test", "Test", changes_request))

    def test_button(self):
        button = Button("Test", True)
        self.assertEqual(str(button), "- [x] Test")

    def test_button_list(self):
        buttons = [Button("Test1", False), Button("Test2", True)]
        button_list = ButtonList("## My Buttons", buttons)
        self.assertEqual(
            button_list.serialize(), "## My Buttons\n- [ ] Test1\n- [x] Test2"
        )
        self.assertEqual(button_list.get_clicked_buttons(), [buttons[1]])


if __name__ == "__main__":
    unittest.main()



import re

from loguru import logger

from sweepai.core.chat import ChatGPT
from sweepai.core.entities import Message
from sweepai.utils.diff import generate_diff

response_format = """Respond using the following structured format:

<judgement_on_task>
Provide extensive, highly detailed criteria for evaluating the contractor's performance, such as:
- Did they identify every single relevant file needed to solve the issue, including all transitive dependencies?
- Did they use multiple code/function/class searches to exhaustively trace every usage and dependency of relevant classes/functions?
- Did they justify why each file is relevant and needed to solve the issue?
- Did they demonstrate a complete, comprehensive understanding of the entire relevant codebase and architecture?

Go through the contractor's process step-by-step. For anything they did even slightly wrong or non-optimally, call it out and explain the correct approach. Be extremely harsh and scrutinizing. If they failed to use enough code/function/class searches to find 100% of relevant usages or if they missed any files that are needed, point these out as critical mistakes. Do not give them the benefit of the doubt on anything.
</judgement_on_task>

<overall_score>
Evaluate the contractor from 1-10, erring on the low side:
1 - Completely failed to identify relevant files, trace dependencies, or understand the issue
2 - Identified a couple files from the issue description but missed many critical dependencies 
3 - Found some relevant files but had major gaps in dependency tracing and codebase understanding
4 - Identified several key files but still missed important usages and lacked justification
5 - Found many relevant files but missed a few critical dependencies
6 - Identified most key files and dependencies but still had some gaps in usage tracing
7 - Found nearly all relevant files but missed a couple edge case usages or minor dependencies
8 - Exhaustively traced nearly all dependencies with robust justification, only minor omissions
9 - Perfectly identified every single relevant file and usage with airtight justification 
10 - Flawless, absolutely exhaustive dependency tracing and codebase understanding
</overall_score>

<message_to_contractor>
Provide a single sentence of extremely specific, targeted, and actionable critical feedback, addressed directly to the contractor.
9-10: Flawless work exhaustively using code/function/class searches to identify 100% of necessary files and usages!
5-8: You failed to search for [X, Y, Z] to find all usages of [class/function]. You need to understand [A, B, C] dependencies.
1-4: You need to search for [X, Y, Z] classes/functions to find actually relevant files. You missed [A, B, C] critical dependencies completely.
</message_to_contractor>

Do not give any positive feedback unless the contractor literally achieved perfection. Be extremely harsh and critical in your evaluation. Assume incompetence until proven otherwise. Make the contractor work hard to get a high score."""

state_eval_prompt = """You are helping contractors on a task that involves finding all of the relevant files needed to resolve a github issue. You are an expert at this task and have solved it hundreds of times. This task does not involve writing or modifying code. The contractors' goal is to identify all necessary files, not actually implement the solution. The contractor should not be coding at all. 

Your job is to review the contractor's work with an extremely critical eye. Leave no stone unturned in your evaluation. Read through every single step the contractor took and analyze it in depth.

""" + response_format + \
"""
Here are some examples of how you should evaluate the contractor's work:

<examples>
Example 1 (Score: 9):
<judgement_on_task>
The contractor did an outstanding job identifying all of the relevant files needed to resolve the payment processing issue. They correctly identified the core Payment.java model where the payment data is defined, and used extensive code searches for "Payment", "pay", "process", "transaction", etc. to exhaustively trace every single usage and dependency.

They found the PaymentController.java and PaymentService.java files where Payment objects are created and processed, and justified how these are critical for the payment flow. They also identified the PaymentRepository.java DAO that interacts with the payments database.

The contractor demonstrated a deep understanding of the payment processing architecture by tracing the dependencies of the PaymentService on external payment gateways like StripeGateway.java and PayPalGateway.java. They even found the PaymentNotificationListener.java that handles webhook events from these gateways.

To round out their analysis, the contractor identified the PaymentValidator.java and PaymentSecurityFilter.java as crucial parts of the payment processing pipeline for validation and security. They justified the relevance of each file with clear explanations tied to the reported payment bug.

No relevant files seem to have been missed. The contractor used a comprehensive set of searches for relevant classes, functions, and terms to systematically map out the entire payment processing codebase. Overall, this shows an excellent understanding of the payment architecture and all its nuances.
</judgement_on_task>
<overall_score>9</overall_score>
<message_to_contractor>
Excellent work identifying Payment.java, PaymentController.java, PaymentService.java, and all critical dependencies.
</message_to_contractor>

Example 2 (Score: 4): 
<judgement_on_task>
The contractor identified the UserAccount.java file where the login bug is occurring, but failed to use nearly enough code/function/class searches to find many other critical files. While they noted that LoginController.java calls UserAccount.authenticateUser(), they didn't search for the "authenticateUser" function to identify LoginService.java which orchestrates the login flow.  

They completely missed using searches for the "UserAccount" class, "credentials", "principal", "login", etc. to find the UserRepository.java file that loads user data from the database and many other files involved in authentication. Searching for "hash", "encrypt", "password", etc. should have revealed the critical PasswordEncryptor.java that handles password hashing.

The contractor claimed UserForgotPasswordController.java and UserCreateController.java are relevant, but failed to justify this at all. These files are not directly related to the login bug.

In general, the contractor seemed to stumble upon a couple relevant files, but failed to systematically trace the login code path and its dependencies. They showed a superficial and incomplete understanding of the login architecture and process. Many critical files were completely missed and the scope was not properly focused on login.
</judgement_on_task>
<overall_score>4</overall_score>  
<message_to_contractor>
Failed to search for "authenticateUser", "UserAccount", "login", "credentials". Missed LoginService.java, UserRepository.java, PasswordEncryptor.java.
</message_to_contractor>

Example 3 (Score: 2):
<judgement_on_task>
The files identified by the contractor, like index.html, styles.css, and ProductList.vue, are completely irrelevant for resolving the API issue with product pricing. The front-end product list display code does not interact with the pricing calculation logic whatsoever.

The contractor completely failed to focus their investigation on the backend api/products/ directory where the pricing bug actually occurs. They did not perform any searches for relevant classes/functions like "Product", "Price", "Discount", etc. to find the ProductController.java API endpoint and the PriceCalculator.java service it depends on.

Basic searches for the "Product" class should have revealed the Product.java model and ProductRepository.java database access code as highly relevant, but these were missed. The contractor failed to demonstrate any understanding of the API architecture and the flow of pricing data from the database to the API response.

The contractor also did not look for any configuration files that provide pricing data, which would be critical for the pricing calculation. They did not search for "price", "cost", etc. in JSON or properties files.

Overall, the contractor seemed to have no clue about the actual pricing bug or the backend API codebase. They looked in completely the wrong places, failed to perform any relevant code/function/class searches, and did not identify a single relevant file for the reported bug. This shows a fundamental lack of understanding of the pricing feature and backend architecture.
</judgement_on_task>
<overall_score>2</overall_score>
<message_to_contractor>
index.html, styles.css, ProductList.vue are irrelevant. Search api/products/ for "Product", "Price", "Discount" classes/functions.
</message_to_contractor>

Example 4 (Score: 7):
<judgement_on_task>
The contractor identified most of the key files involved in the user profile update process, including UserProfileController.java, UserProfileService.java, and UserProfile.java. They correctly traced the flow of data from the API endpoint to the service layer and model.

However, they missed a few critical dependencies. They did not search for "UserProfile" to find the UserProfileRepository.java DAO that loads and saves user profiles to the database. This is a significant omission in their understanding of the data persistence layer.

The contractor also failed to look for configuration files related to user profiles. Searching for "profile" in YAML or properties files should have revealed application-profiles.yml which contains important profile settings. 

While the contractor had a decent high-level understanding of the user profile update process, they showed some gaps in their low-level understanding of the data flow and configuration. They needed to be more thorough in tracing code dependencies to uncover the complete set of relevant files.
</judgement_on_task>
<overall_score>7</overall_score>
<message_to_contractor>
Missed UserProfileRepository.java and application-profiles.yml dependencies. Search for "UserProfile" and "profile" to find remaining relevant files.
</message_to_contractor>
</examples>"""

modify_eval_response_format = """Please provide your critical evaluation of this submission using the following structured format:

<judgement_on_task>
Your judgement here explaining in detail how you are evaluating the contractor's work against the original requirements. Call out specific issues, gaps, or places where the contractor went wrong. Provide clear justification and examples to support your assessment.
</judgement_on_task>

<overall_score>
Evaluate the contractor from 1-10, erring on the low side:
1 - No attempt made to complete the task
2 - Minimal attempt with little to no functional code changes
3 - Significant gaps in completion of the task, code largely non-functional
4 - Partial completion of the task with major issues and errors
5 - Task partially satisfied but with significant issues remaining
6 - Most requirements addressed but some notable gaps or issues
7 - Task mostly satisfied with a few minor issues or improvements needed
8 - Task fully satisfied with working code, minor improvements possible
9 - Task fully satisfied with high-quality, efficient, and maintainable code
10 - Superhuman completion of the task, exceptional code quality and design
</overall_score>

<message_to_contractor>
Provide 3-5 specific, actionable pieces of feedback here for the contractor to focus on in their next attempt. For example:

1. Import the missing XYZ library at the top of file A to avoid compilation errors.
2. The FooBar() method called on line 127 of file B is not defined anywhere. Implement this method or remove the call.
3. The current changes do not handle the edge case of X. Add logic to check for this case and respond appropriately.
4. Consider refactoring the code in function ABC to be more readable and maintainable. It is currently very complex and difficult to follow.

Focus on the most critical issues that are blocking functional completion of the task first, then cover code quality and best practices.
</message_to_contractor>"""

modify_eval_examples = """Example 1:

<judgement_on_task>
The contractor has done an exceptional job completing the task of optimizing the database queries and improving the overall performance of the application. They accurately identified the bottlenecks in the existing code and made targeted, efficient changes to address them. The updated queries now utilize appropriate indexes, avoid unnecessary joins, and minimize data transfer. The contractor also added helpful comments explaining their optimization strategies, making the code more maintainable. All the original requirements have been fully satisfied, and the code changes demonstrate a deep understanding of database performance best practices.
</judgement_on_task>

<overall_score>9</overall_score>

<message_to_contractor>
Great work on this optimization task! Your code changes have significantly improved the efficiency of the database queries. A few minor suggestions for further enhancement:

1. Consider using a parameterized query on line 95 of `queries.py` to avoid potential SQL injection vulnerabilities.
2. The `get_user_data` function in `utils.py` could benefit from some additional error handling to gracefully deal with potential edge cases, such as a user ID not being found.

Overall, this is a high-quality submission. Keep up the excellent work!
</message_to_contractor>

Example 2:

<judgement_on_task>
The contractor has made an attempt to implement the new feature for generating PDF reports, but there are several gaps and issues in their code changes. While they have correctly added a new endpoint for triggering the report generation, the actual PDF creation logic is incomplete. The code is currently missing the necessary imports for the PDF library, and there are several undefined variables and functions being used. Additionally, the error handling is insufficient, which could lead to uncaught exceptions. The contractor has also not included any unit tests to verify the functionality of the new feature. More work is needed to fully satisfy the requirements and ensure a reliable, maintainable solution.
</judgement_on_task>

<overall_score>5</overall_score>

<message_to_contractor>
Thank you for your efforts on implementing the PDF report feature. However, there are several areas that need improvement:

1. Add the necessary imports for the PDF library at the top of `report_generator.py`. Currently, the `import pdf_lib` statement is missing.
2. Implement the missing `generate_pdf` function that is currently being called on line 42 of `report_generator.py`. This function should contain the core logic for creating the PDF report.
3. Fix the undefined variables `report_data` and `user_id` in the `generate_report` endpoint. Ensure that these variables are properly initialized before being used.
4. Add proper error handling to the `generate_report` endpoint to catch and handle any exceptions that may occur during the PDF generation process.
5. Write unit tests for the new feature to verify its functionality and catch any potential bugs.

Please address these issues and resubmit your code changes for further review.
</message_to_contractor>

Example 3:

<judgement_on_task>
The contractor's submission for the task of implementing a new user authentication system is severely lacking and does not meet the requirements. The code changes are minimal and do not include any of the core functionality needed for user registration, login, or password hashing. The contractor has merely added a few empty functions and commented out some existing code, without any actual implementation. There are no changes made to the database schema to support storing user credentials securely. The submission also introduces several syntax errors and undefined variables, indicating a lack of basic coding proficiency. Overall, this submission falls far short of the expected solution and does not demonstrate any meaningful progress towards completing the task.
</judgement_on_task>

<overall_score>2</overall_score>

<message_to_contractor>
I regret to inform you that your submission for the user authentication task is unacceptable and requires significant improvement. The following critical issues must be addressed:

1. Implement the core functionality for user registration, including validating input data, securely storing user credentials in the database, and handling duplicate username/email scenarios.
2. Add the necessary code for user login, including verifying the provided credentials against the stored data and generating a secure authentication token upon successful login.
3. Integrate a secure password hashing algorithm, such as bcrypt or scrypt, to store and compare passwords instead of storing them in plain text.
4. Update the database schema to include the required tables and fields for storing user information and authentication data.
5. Fix all syntax errors and undefined variables in your code. Ensure that your code is free of basic compilation errors before submitting.

I recommend reviewing the task requirements carefully, studying best practices for user authentication, and taking the time to implement a complete and secure solution. If you need further guidance or clarification, please don't hesitate to ask.
</message_to_contractor>"""

modify_eval_prompt = """You are an evaluator agent tasked with grading and providing critical feedback on code changes submitted by an outside contractor in response to a given coding task. You will be provided with the original task description as well as a series of file changes in unified diff format.

Your job is to carefully review the code changes and provide feedback focused on the following:

1. Identify any missing import statements that would prevent the code from compiling. Call out the specific imports that are needed.

2. Look for any variables or methods that are referenced but not defined in the provided code changes. These may indicate the contractor hallucinated or made invalid assumptions. 

3. Analyze whether the code changes, as submitted, fully satisfy the requirements of the original coding task. Identify any gaps or ways in which the solution falls short.

Remember, your goal is to be a harsh critic and really scrutinize the work to ensure only high-quality, complete code changes are accepted. Do not praise mediocre work.

""" + modify_eval_response_format + modify_eval_examples

modify_eval_patch_prompt = """\
You are a meticulous code reviewer providing critical and specific feedback on a contractor's code changes to help resolve a GitHub issue.
Inputs:
- Task description
- Code patch (diff) 
- Completed changes
- Current plan
- Current file
Steps:
1. Review CURRENT TASK for requirements.
2. Analyze code patch:
   - Purpose and impact of each change
   - Check for LLM failures: 
     - Logic errors
     - Unhandled edge cases
     - Missing imports
     - Incomplete changes
     - Undefined variables/functions
     - Usage of nullable attributes
     - Non-functional code
   - Alignment with plan and requirements
3. Perform critical contextual analysis:
   - Break down changes 
   - Explain reasoning
   - Identify logic issues, edge cases, plan deviations
   - Consider all scenarios and pitfalls
   - Consider backwards compatibility and future-proofing
   - Suggest fixes for problems
4. Be extremely critical. Do not overlook ANY issues.
Format:
<task_summary>
Provide a brief summary of the task requirements, the contractor's plan, and the current file changes.
</task_summary>
<patch_integration>
Critically analyze patch fit, behavior changes, conflicts, issues, consequences. 
</patch_integration>
<code_examination>
Break down changes. Explain purpose. Call out logic errors and integration issues in detail:
- Unhandled edge cases: [list]
- Logic errors: [list]
- Missing imports: [list]
- Incomplete changes: [list] 
- Undefined variables/functions: [list]
- Non-functional code: [list]
Require justification for plan deviations. Criticize behavior changes not handled. Overlook NOTHING.
</code_examination>
<feedback>
Give critical, specific feedback on logic and integration ONLY. LIMIT FEEDBACK TO CURRENT TASK'S SCOPE. NO EXTRA SUGGESTIONS.
</feedback>
<next_step>
COMPLETE - mark the CURRENT TASK as complete
CONTINUE - apply the current changes, but make additional fixes before marking the CURRENT TASK as complete
REJECT - generate the code again
</next_step>

Focus on functional changes, logic errors and other issues. Do not provide feedback on code style,comments or docstrings unless they're necessary."""

modify_eval_suffix_prompt = """Again, you will critically review the code changes and consider the following concerns and respond in the following format. Your feedback will be very specific.

Inputs:
- Task description
- Code patch (diff) 
- Completed changes
- Current plan
- Current file
Steps:
1. Review CURRENT TASK for requirements.
2. Analyze code patch:
   - Purpose and impact of each change
   - Check for LLM failures: 
     - Logic errors
     - Unhandled edge cases
     - Missing imports
     - Incomplete changes
     - Undefined variables/functions
     - Usage of nullable attributes
     - Non-functional code
   - Alignment with plan and requirements
3. Perform critical contextual analysis:
   - Break down changes 
   - Explain reasoning
   - Identify logic issues, edge cases, plan deviations
   - Consider all scenarios and pitfalls
   - Consider backwards compatibility and future-proofing
   - Suggest fixes for problems
   - Evaluate error handling and fallback mechanisms
4. Be extremely critical. Do not overlook ANY issues.

Format:

<patch_integration>
Critically analyze patch fit, behavior changes, conflicts, issues, consequences. 
</patch_integration>

<code_examination>
Break down changes. Explain purpose. Call out logic errors and integration issues in detail:
- Unhandled edge cases: [list]
- Logic errors: [list]
- Missing imports: [list]
- Incomplete changes: [list] 
- Undefined variables/functions: [list]
- Non-functional code: [list]
Require justification for plan deviations. Criticize behavior changes not handled. Overlook NOTHING.
</code_examination>

<feedback>
Give critical, specific feedback on logic and integration ONLY. LIMIT FEEDBACK TO THE SCOPE OF THE CURRENT TASK. NO EXTRA SUGGESTIONS.
</feedback>

<next_step>
REJECT - the code is a step backwards, so we should revert the patch and generate the code again
CONTINUE - apply the current changes, but make additional tweaks before moving on to the next task of the plan
COMPLETE - mark the CURRENT TASK as complete as there are no concerns or missed edge cases
</next_step>

Note: Only mark the task as complete if you are confident that all requirements have been met, edge cases have been handled, error handling and fallback mechanisms are in place, and no further specific improvements are necessary. If there are any specific doubts or actionable suggestions for enhancements, provide feedback and mark the task as "CONTINUE". Again, limit the feedback to the scope of the current task.

Focus on functional changes, logic errors and other issues. Do not provide feedback on code style,comments or docstrings unless they're necessary.

Respond with your extremely critical analysis and feedback."""

# general framework for a dfs search
# 1. sample trajectory
# 2. for each trajectory, run the assistant until it hits an error or end state
#    - in either case perform self-reflection
# 3. update the reflections section with the new reflections
CLAUDE_MODEL = "claude-3-opus-20240229"

class EvaluatorAgent(ChatGPT):
    def evaluate_run(self, problem_statement: str, run_text: str, stored_files: list[str]):
        self.model = CLAUDE_MODEL
        self.messages = [Message(role="system", content=state_eval_prompt)]
        formatted_problem_statement = f"This is the task for the contractor to research:\n<task_to_research>\n{problem_statement}\n</task_to_research>"
        contractor_stored_files = "\n".join([file for file in stored_files])
        stored_files_section = f"""The contractor stored these files:\n<stored_files>\n{contractor_stored_files}\n</stored_files>"""
        content = formatted_problem_statement + "\n\n" + f"<contractor_attempt>\n{run_text}\n</contractor_attempt>"\
             + f"\n\n{stored_files_section}\n\n" + response_format
        evaluate_response = self.chat_anthropic(
            content=content,
            stop_sequences=["</message_to_contractor>"],
            model=CLAUDE_MODEL,
            message_key="user_request",
        )
        evaluate_response += "</message_to_contractor>" # add the stop sequence back in, if it stopped for another reason we've crashed
        overall_score = None
        message_to_contractor = None
        try:
            overall_score_pattern = r"<overall_score>(.*?)</overall_score>"
            message_to_contractor_pattern = r"<message_to_contractor>(.*?)</message_to_contractor>"

            overall_score_match = re.search(overall_score_pattern, evaluate_response, re.DOTALL)
            message_to_contractor_match = re.search(message_to_contractor_pattern, evaluate_response, re.DOTALL)

            if overall_score_match is None or message_to_contractor_match is None:
                return overall_score, message_to_contractor

            overall_score = overall_score_match.group(1).strip()
            # check if 1 through 10 are a match
            if not re.match(r"^[1-9]|10$", overall_score):
                return None, None
            else:
                overall_score_match = re.match(r"^[1-9]|10$", overall_score)
                overall_score = overall_score_match.group(0).strip()
            overall_score = int(overall_score)
            message_to_contractor = message_to_contractor_match.group(1).strip()
            return overall_score, message_to_contractor
        except Exception as e:
            logger.info(f"Error evaluating response: {e}")
            return overall_score, message_to_contractor

# Eval agent specific to modify step
class ModifyEvaluatorAgent(ChatGPT):
    def evaluate_patch(
        self, 
        problem_statement: str, 
        patch: str, 
        changed_files: dict[str, dict[str, str]], 
        new_file_contents: str,
        current_plan: str, 
        current_task: str,
        file_name: str,
        warning_message: str = "",
        previous_attempt: str = "",
        chat_logger_messages: list[dict[str, str]] | None = None
    ):
        self.model = CLAUDE_MODEL
        self.messages = [Message(role="system", content=modify_eval_patch_prompt)]
        formatted_problem_statement = f"This is the task for the contractor to complete:\n<task_to_complete>\n{problem_statement}\n</task_to_complete>\n\n"
        formatted_patch_and_contents = f"This is the CURRENT PATCH that the contractor has submitted for evaluation:\n<current_patch file_name={file_name}>\n{patch}\n</current_patch>\n\n" + f"This is the current file after modifications:\n<current_file>\n{new_file_contents}\n</current_file>\n\n"
        formatted_plan = f"This is the current plan that we must follow:\n<entire_plan>\n{current_plan}\n</entire_plan>\n\n"
        contractor_changes_made: dict[str, str] = {}
        for file_name, file_data in changed_files.items():
            if "original_contents" not in file_data or "contents" not in file_data:
                continue
            diff = generate_diff(file_data["original_contents"], file_data["contents"])
            if diff:
                contractor_changes_made[file_name] = diff
        contractor_changed_files = "\n".join([f"<completed_patch file_name={file_name}>\n{diff}\n</completed_patch>" for file_name, diff in contractor_changes_made.items()])
        changed_files_section = f"""The contractor has already completed these changes as part of the completed tasks:\n<completed_changes>\n{contractor_changed_files}\n</completed_changes>\n\n"""
        content = formatted_problem_statement + formatted_plan + changed_files_section + formatted_patch_and_contents
        if warning_message:
            content += f"The changes also trigger the following warnings:\n<warnings>\n{warning_message}\n</warnings>\n\n"
        content += current_task
        if previous_attempt:
            content += "\n\n" + previous_attempt
        content += "\n\n" + modify_eval_suffix_prompt
        # if warning_message:
        #     breakpoint()
        evaluate_response = self.chat_anthropic(
            content=content,
            stop_sequences=["</message_to_contractor>"],
            model=CLAUDE_MODEL,
            message_key="user_request",
        )
        evaluate_response += "</message_to_contractor>" # add the stop sequence back in, if it stopped for another reason we've crashed
        # breakpoint()
        # update chat_logger_messages in place if they are passed in
        if chat_logger_messages:
            chat_logger_messages.append({"role": "assistant", "content": content})
            chat_logger_messages.append({"role": "user", "content": evaluate_response})
        next_step = None
        feedback = ""
        try:
            next_step_pattern = r"<next_step>(.*?)</next_step>"
            message_to_contractor_pattern = r"<feedback>(.*?)</feedback>"

            next_step_match = re.search(next_step_pattern, evaluate_response, re.DOTALL)
            message_to_contractor_match = re.search(message_to_contractor_pattern, evaluate_response, re.DOTALL)

            if next_step_match is None or message_to_contractor_match is None:
                return next_step, feedback

            next_step = next_step_match.group(1).strip()
            # check if 1 through 10 are a match
            if not any(["COMPLETE" in next_step, "CONTINUE" in next_step, "REJECT" in next_step]):
                return None, ""
            else:
                if "COMPLETE" in next_step:
                    next_step = "COMPLETE"
                elif "CONTINUE" in next_step:
                    next_step = "CONTINUE"
                else:
                    next_step = "REJECT"
            feedback = message_to_contractor_match.group(1).strip()
            return next_step, feedback
        except Exception as e:
            logger.info(f"Error evaluating response: {e}")
            return next_step, feedback


    def evaluate_run(self, problem_statement: str, run_text: str, changed_files: dict[str, dict[str, str]]):
        self.model = CLAUDE_MODEL
        self.messages = [Message(role="system", content=modify_eval_prompt)]
        formatted_problem_statement = f"This is the task for the contractor to complete:\n<task_to_complete>\n{problem_statement}\n</task_to_complete>"
        contractor_changes_made: dict[str, str] = {}
        for file_name, file_data in changed_files.items():
            diff = generate_diff(file_data["original_contents"], file_data["contents"])
            if diff:
                contractor_changes_made[file_name] = diff
        contractor_changed_files = "\n".join([f"Changes made to file {file_name}:\n\n{diff}\n\n" for file_name, diff in contractor_changes_made.items()])
        changed_files_section = f"""The contractor made these changes to the following files:\n<changed_files>\n{contractor_changed_files}\n</changed_files>"""
        content = formatted_problem_statement + "\n\n" + f"<contractor_attempt>\n{run_text}\n</contractor_attempt>"\
             + f"\n\n{changed_files_section}\n\n" + modify_eval_response_format
        evaluate_response = self.chat_anthropic(
            content=content,
            stop_sequences=["</message_to_contractor>"],
            model=CLAUDE_MODEL,
            message_key="user_request",
        )
        evaluate_response += "</message_to_contractor>" # add the stop sequence back in, if it stopped for another reason we've crashed
        overall_score = None
        message_to_contractor = None
        try:
            overall_score_pattern = r"<overall_score>(.*?)</overall_score>"
            message_to_contractor_pattern = r"<message_to_contractor>(.*?)</message_to_contractor>"

            overall_score_match = re.search(overall_score_pattern, evaluate_response, re.DOTALL)
            message_to_contractor_match = re.search(message_to_contractor_pattern, evaluate_response, re.DOTALL)

            if overall_score_match is None or message_to_contractor_match is None:
                return overall_score, message_to_contractor

            overall_score = overall_score_match.group(1).strip()
            # check if 1 through 10 are a match
            if not re.match(r"^[1-9]|10$", overall_score):
                return None, None
            else:
                overall_score_match = re.match(r"^[1-9]|10$", overall_score)
                overall_score = overall_score_match.group(0).strip()
            overall_score = int(overall_score)
            message_to_contractor = message_to_contractor_match.group(1).strip()
            return overall_score, message_to_contractor
        except Exception as e:
            logger.info(f"Error evaluating response: {e}")
            return overall_score, message_to_contractor

if __name__ == "__main__":
    try:
        pass
    except Exception as e:
        import sys
        info = sys.exc_info()
        import pdb
        # pylint: disable=no-member
        pdb.post_mortem(info[2])
        raise e


"""
on_comment is responsible for handling PR comments and PR review comments, called from sweepai/api.py.
It is also called in sweepai/handlers/on_ticket.py when Sweep is reviewing its own PRs.
"""
import re
import time
import traceback
from typing import Any

from loguru import logger
from tabulate import tabulate

from sweepai.config.client import get_blocked_dirs, get_documentation_dict
from sweepai.config.server import (
    DEFAULT_GPT4_32K_MODEL,
    ENV,
    GITHUB_BOT_USERNAME,
    MONGODB_URI,
)
from sweepai.core.context_pruning import get_relevant_context
from sweepai.core.entities import FileChangeRequest, MockPR, NoFilesException
from sweepai.core.sweep_bot import SweepBot, get_files_to_change, validate_file_change_requests
from sweepai.handlers.on_review import get_pr_diffs
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import ClonedRepo, get_github_client
from sweepai.utils.progress import TicketProgress
from sweepai.utils.prompt_constructor import HumanMessageCommentPrompt
from sweepai.utils.str_utils import BOT_SUFFIX, FASTER_MODEL_MESSAGE
from sweepai.utils.ticket_utils import fire_and_forget_wrapper, prep_snippets

num_of_snippets_to_query = 30
total_number_of_snippet_tokens = 15_000
num_full_files = 2
num_extended_snippets = 2

ERROR_FORMAT = "❌ {title}\n\nPlease join our [Discord](https://discord.gg/sweep) to report this issue."


def on_comment(
    repo_full_name: str,
    repo_description: str,
    comment: str,
    pr_path: str | None,
    pr_line_position: int | None,
    username: str,
    installation_id: int,
    pr_number: int = None,
    comment_id: int | None = None,
    chat_logger: Any = None,
    pr: MockPR = None,  # For on_comment calls before PR is created
    repo: Any = None,
    comment_type: str = "comment",
    type: str = "comment",
    tracking_id: str = None,
):
    with logger.contextualize(
        tracking_id=tracking_id,
    ):
        logger.info(
            f"Calling on_comment() with the following arguments: {comment},"
            f" {repo_full_name}, {repo_description}, {pr_path}"
        )
        organization, repo_name = repo_full_name.split("/")
        start_time = time.time()

        _token, g = get_github_client(installation_id)
        repo = g.get_repo(repo_full_name)
        if pr is None:
            pr = repo.get_pull(pr_number)
        pr_title = pr.title
        pr_body = (
            pr.body.split("<details>\n<summary><b>🎉 Latest improvements to Sweep:")[0]
            if pr.body
            and "<details>\n<summary><b>🎉 Latest improvements to Sweep:" in pr.body
            else pr.body
        )
        pr_file_path = None
        diffs = get_pr_diffs(repo, pr)
        pr_chunk = None
        formatted_pr_chunk = None

        assignee = pr.assignee.login if pr.assignee else None
        issue_number_match = re.search(r"Fixes #(?P<issue_number>\d+).", pr_body or "")
        original_issue = None
        if issue_number_match or assignee:
            issue_number = issue_number_match.group("issue_number")
            if not assignee:
                # issue_number = issue_number_match.group("issue_number")
                original_issue = repo.get_issue(int(issue_number))
                author = original_issue.user.login
            else:
                author = assignee
            logger.info(f"Author of original issue is {author}")
            chat_logger = (
                chat_logger
                if chat_logger is not None
                else ChatLogger(
                    {
                        "repo_name": repo_name,
                        "title": "(Comment) " + pr_title,
                        "issue_url": pr.html_url,
                        "pr_file_path": pr_file_path,  # may be None
                        "pr_chunk": pr_chunk,  # may be None
                        "repo_full_name": repo_full_name,
                        "repo_description": repo_description,
                        "comment": comment,
                        "pr_path": pr_path,
                        "pr_line_position": pr_line_position,
                        "username": author,
                        "installation_id": installation_id,
                        "pr_number": pr_number,
                        "type": "comment",
                    },
                    active=True,
                )
                if MONGODB_URI
                else None
            )
        else:
            chat_logger = None

        if chat_logger:
            is_paying_user = chat_logger.is_paying_user()
            use_faster_model = chat_logger.use_faster_model()
        else:
            # Todo: chat_logger is None for MockPRs, which will cause all comments to use GPT-4
            is_paying_user = True
            use_faster_model = False
        
        if use_faster_model:
            raise Exception(FASTER_MODEL_MESSAGE)

        assignee = pr.assignee.login if pr.assignee else None

        metadata = {
            "repo_full_name": repo_full_name,
            "repo_name": repo_name,
            "organization": organization,
            "repo_description": repo_description,
            "installation_id": installation_id,
            "username": username if not username.startswith("sweep") else assignee,
            "function": "on_comment",
            "model": "gpt-4",
            "tier": "pro" if is_paying_user else "free",
            "mode": ENV,
            "pr_path": pr_path,
            "pr_line_position": pr_line_position,
            "pr_number": pr_number or pr.id,
            "pr_html_url": pr.html_url,
            "comment_id": comment_id,
            "comment": comment,
            "issue_number": issue_number if issue_number_match else "",
            "tracking_id": tracking_id,
        }

        logger.bind(**metadata)

        elapsed_time = time.time() - start_time
        posthog.capture(
            username,
            "started",
            properties={
                **metadata,
                "duration": elapsed_time,
                "tracking_id": tracking_id,
            },
        )
        logger.info(f"Getting repo {repo_full_name}")
        file_comment = bool(pr_path) and bool(pr_line_position)

        item_to_react_to = None
        reaction = None

        bot_comment = None

        def edit_comment(new_comment: str) -> None:
            if bot_comment is not None:
                bot_comment.edit(new_comment + BOT_SUFFIX)

        try:
            # Check if the PR is closed
            if pr.state == "closed":
                return {"success": True, "message": "PR is closed. No event fired."}
            if comment_id:
                try:
                    item_to_react_to = pr.get_issue_comment(comment_id)
                    reaction = item_to_react_to.create_reaction("eyes")
                except Exception:
                    try:
                        item_to_react_to = pr.get_review_comment(comment_id)
                        reaction = item_to_react_to.create_reaction("eyes")
                    except Exception:
                        pass

                if reaction is not None:
                    # Delete rocket reaction
                    reactions = item_to_react_to.get_reactions()
                    for r in reactions:
                        if (
                            r.content == "rocket"
                            and r.user.login == GITHUB_BOT_USERNAME
                        ):
                            item_to_react_to.delete_reaction(r.id)

            branch_name = (
                pr.head.ref if pr_number else pr.pr_head  # pylint: disable=no-member
            )
            cloned_repo = ClonedRepo(
                repo_full_name,
                installation_id,
                branch=branch_name,
                repo=repo,
                token=_token,
            )

            # Generate diffs for this PR
            pr_diff_string = None
            pr_files_modified = None
            if pr_number:
                patches = []
                pr_files_modified = {}
                files = pr.get_files()
                for file in files:
                    if file.status == "modified":
                        # Get the entire file contents, not just the patch
                        pr_files_modified[file.filename] = repo.get_contents(
                            file.filename, ref=branch_name
                        ).decoded_content.decode("utf-8")

                        patches.append(
                            f'<file file_path="{file.filename}">\n{file.patch}\n</file>'
                        )
                pr_diff_string = (
                    "<files_changed>\n" + "\n".join(patches) + "\n</files_changed>"
                )

            # This means it's a comment on a file
            if file_comment:
                pr_file = repo.get_contents(
                    pr_path, ref=branch_name
                ).decoded_content.decode("utf-8")
                pr_lines = pr_file.splitlines()
                start = max(0, pr_line_position - 11)
                end = min(len(pr_lines), pr_line_position + 10)
                original_line = pr_lines[pr_line_position - 1]
                pr_chunk = "\n".join(pr_lines[start:end])
                pr_file_path = pr_path.strip()
                formatted_pr_chunk = (
                    "\n".join(pr_lines[start : pr_line_position - 1])
                    + f"\n{pr_lines[pr_line_position - 1]} <- COMMENT: {comment.strip()}\n"
                    + "\n".join(pr_lines[pr_line_position:end])
                )
                if comment_id:
                    try:
                        bot_comment = pr.create_review_comment_reply(
                            comment_id, "Working on it..." + BOT_SUFFIX
                        )
                    except Exception as e:
                        print(e)
            else:
                formatted_pr_chunk = None  # pr_file
                bot_comment = pr.create_issue_comment("Working on it..." + BOT_SUFFIX)
            if file_comment:
                snippets = []
                tree = ""
            else:
                try:
                    search_query = (comment).strip("\n")
                    formatted_query = (f"{comment}").strip("\n")
                    repo_context_manager = prep_snippets(
                        cloned_repo, search_query, TicketProgress(tracking_id="none")
                    )
                    repo_context_manager = get_relevant_context(
                        formatted_query,
                        repo_context_manager,
                        TicketProgress(tracking_id="none"),
                        chat_logger=chat_logger,
                    )
                    snippets = repo_context_manager.current_top_snippets
                    tree = str(repo_context_manager.dir_obj)
                    cloned_repo = repo_context_manager.cloned_repo
                except Exception as e:
                    logger.error(traceback.format_exc())
                    raise e
            get_documentation_dict(repo)
            docs_results = ""
            logger.info("Getting response from ChatGPT...")
            human_message = HumanMessageCommentPrompt(
                comment=comment,
                repo_name=repo_name,
                repo_description=repo_description if repo_description else "",
                diffs=diffs,
                title=pr_title,
                tree=tree,
                summary=pr_body,
                snippets=snippets,
                # commit_history=commit_history,
                pr_file_path=pr_file_path,  # may be None
                pr_chunk=formatted_pr_chunk,  # may be None
                original_line=original_line if pr_chunk else None,
                relevant_docs=docs_results,
            )
            logger.info(f"Human prompt{human_message.construct_prompt()}")

            if use_faster_model:
                raise Exception("GPT-3.5 is not supported for comments")
            sweep_bot = SweepBot.from_system_message_content(
                # human_message=human_message, model="claude-v1.3-100k", repo=repo
                human_message=human_message,
                repo=repo,
                chat_logger=chat_logger,
                model=DEFAULT_GPT4_32K_MODEL,
                cloned_repo=cloned_repo,
            )
        except Exception as e:
            logger.error(traceback.format_exc())
            elapsed_time = time.time() - start_time
            posthog.capture(
                username,
                "failed",
                properties={
                    "error": str(e),
                    "reason": "Failed to get files",
                    "duration": elapsed_time,
                    "tracking_id": tracking_id,
                    **metadata,
                },
            )
            edit_comment(ERROR_FORMAT.format(title="Failed to get files"))
            raise e

        try:
            logger.info("Fetching files to modify/create...")
            if file_comment:
                file_change_requests = [
                    FileChangeRequest(
                        filename=pr_file_path,
                        instructions=f"The user left this comment {comment} in this chunk of code:\n<review_code_chunk>\n{formatted_pr_chunk}\n</review_code_chunk>.\nResolve their comment.",
                        change_type="modify",
                        comment_line=pr_line_position + 1,
                    )
                ]
            else:
                file_change_requests, plan = get_files_to_change(
                    relevant_snippets=repo_context_manager.current_top_snippets,
                    read_only_snippets=repo_context_manager.read_only_snippets,
                    problem_statement=formatted_query,
                    repo_name=repo_name,
                    pr_diffs=pr_diff_string
                )
                validate_file_change_requests(file_change_requests, repo_context_manager.cloned_repo)
                file_change_requests = sweep_bot.validate_file_change_requests(
                    file_change_requests, branch=branch_name
                )

            sweep_response = "I couldn't find any relevant files to change."
            if file_change_requests:
                table_message = tabulate(
                    [
                        [
                            f"`{file_change_request.filename}`",
                            file_change_request.instructions_display.replace(
                                "\n", "<br/>"
                            ).replace("```", "\\```"),
                        ]
                        for file_change_request in file_change_requests
                    ],
                    headers=["File Path", "Proposed Changes"],
                    tablefmt="pipe",
                )
                sweep_response = (
                    f"I am making the following changes:\n\n{table_message}"
                )
            quoted_comment = "> " + comment.replace("\n", "\n> ")
            response_for_user = (
                f"{quoted_comment}\n\nHi @{username},\n\n{sweep_response}"
            )
            if pr_number:
                edit_comment(response_for_user)

            blocked_dirs = get_blocked_dirs(sweep_bot.repo)

            sweep_bot.comment_pr_diff_str = pr_diff_string
            sweep_bot.comment_pr_files_modified = pr_files_modified
            changes_made = sum(
                [
                    change_made
                    for _, change_made, _, _ in sweep_bot.change_files_in_github_iterator(
                        file_change_requests, branch_name, blocked_dirs
                    )
                ]
            )
            try:
                if comment_id:
                    if changes_made:
                        response_for_user = "Done."
                    else:
                        response_for_user = 'I wasn\'t able to make changes. This could be due to an unclear request or a bug in my code.\n As a reminder, comments on a file only modify that file. Comments on a PR (at the bottom of the "conversation" tab) can modify the entire PR. Please try again or contact us on [Discord](https://discord.gg/sweep)'
            except Exception as e:
                logger.error(f"Failed to reply to comment: {e}")

            if not isinstance(pr, MockPR):
                if pr.user.login == GITHUB_BOT_USERNAME and pr.title.startswith(
                    "[DRAFT] "
                ):
                    # Update the PR title to remove the "[DRAFT]" prefix
                    pr.edit(title=pr.title.replace("[DRAFT] ", "", 1))

            logger.info("Done!")
        except NoFilesException:
            elapsed_time = time.time() - start_time
            posthog.capture(
                username,
                "failed",
                properties={
                    "error": "No files to change",
                    "reason": "No files to change",
                    "duration": elapsed_time,
                    **metadata,
                },
            )
            edit_comment(ERROR_FORMAT.format(title="Could not find files to change"))
            return {"success": True, "message": "No files to change."}
        except Exception as e:
            logger.error(traceback.format_exc())
            elapsed_time = time.time() - start_time
            posthog.capture(
                username,
                "failed",
                properties={
                    "error": str(e),
                    "reason": "Failed to make changes",
                    "duration": elapsed_time,
                    **metadata,
                },
            )
            edit_comment(ERROR_FORMAT.format(title="Failed to make changes"))
            raise e

        # Delete eyes
        if reaction is not None:
            item_to_react_to.delete_reaction(reaction.id)

        try:
            item_to_react_to = pr.get_issue_comment(comment_id)
            reaction = item_to_react_to.create_reaction("rocket")
        except Exception:
            try:
                item_to_react_to = pr.get_review_comment(comment_id)
                reaction = item_to_react_to.create_reaction("rocket")
            except Exception:
                pass

        if response_for_user is not None:
            try:
                edit_comment(f"## 🚀 Wrote Changes\n\n{response_for_user}")
            except Exception:
                pass

        elapsed_time = time.time() - start_time
        # make async
        fire_and_forget_wrapper(posthog.capture)(
            username,
            "success",
            properties={
                **metadata,
                "tracking_id": tracking_id,
                "duration": elapsed_time,
            },
        )
        return {"success": True}



import unittest

from sweepai.utils.code_tree import CodeTree


class TestCodeTree(unittest.TestCase):
    def test_get_lines_surrounding(self):
        code = """import math

def square_root(x):
    if x < 0:
        raise ValueError("Negative value")
    print(
        "hello world"
    )
    return math.sqrt(x)

print(square_root(4))
"""
        tree = CodeTree.from_code(code)

        # Line inside the 'import' statement
        result = tree.get_lines_surrounding(0)
        expected = (0, 0)
        self.assertEqual(result, expected)

        # Line not present
        result = tree.get_lines_surrounding(100)
        self.assertEqual(result, (100, 100))


if __name__ == "__main__":
    unittest.main()



import unittest
import unittest.mock

from sweepai import api


class TestAPI(unittest.TestCase):
    def setUp(self):
        self.mock_api = unittest.mock.create_autospec(api)

    def test_home(self):
        self.mock_api.home.return_value = "<h2>Sweep Webhook is up and running! To get started, copy the URL into the GitHub App settings' webhook field.</h2>"
        result = self.mock_api.home()
        self.assertEqual(
            result,
            "<h2>Sweep Webhook is up and running! To get started, copy the URL into the GitHub App settings' webhook field.</h2>",
        )
        self.mock_api.home.assert_called_once()

    # Add more test methods as needed for each function in api.py


if __name__ == "__main__":
    unittest.main()



"""
Take a PR and provide an AI generated review of the PR.
"""
from sweepai.utils.event_logger import logger


def get_pr_diffs(repo, pr):
    base_sha = pr.base.sha
    head_sha = pr.head.sha

    comparison = repo.compare(base_sha, head_sha)
    file_diffs = comparison.files

    pr_diffs = []
    for file in file_diffs:
        diff = file.patch
        if (
            file.status == "added"
            or file.status == "modified"
            or file.status == "removed"
        ):
            pr_diffs.append((file.filename, diff))
        else:
            logger.info(
                f"File status {file.status} not recognized"
            )  # TODO(sweep): We don't handle renamed files
    return pr_diffs



"""
We should input an updated snippet generated by Sweep, and output the comment that was left in the code
"""

import re

from loguru import logger

from sweepai.core.chat import ChatGPT
from sweepai.core.entities import Message, RegexMatchableBaseModel
from sweepai.utils.comment_utils import check_comments_presence

system_message_prompt = """You are a genius engineer tasked with identifying any incomplete logic for the following code change.
You have been provided with the relevant metadata to the issue.
Please identify and extract any leftover comments from the new_code. If there are none, return empty tags."""

user_prompt = """# Code
File path: {file_path}
# Original request
{request}

<new_code>
{new_code}
</new_code>

Review new_code and provide your response in the format:
<leftover_comments>
<leftover_comment>
leftover comment verbatim from the new_code that mentions incomplete changes like "rest of code here". these comments are written like instructions or notes
</leftover_comment>
...
</leftover_comments>"""


class LeftoverComments(RegexMatchableBaseModel):
    leftover_comments: list[str] = []

    @classmethod
    def from_string(cls, leftover_comments_response: str, **kwargs):
        leftover_comments = []
        leftover_comments_pattern = (
            r"""<leftover_comment>(\n)?(?P<leftover_comment>.*?)</leftover_comment>"""
        )
        for match_ in re.finditer(
            leftover_comments_pattern, leftover_comments_response, re.DOTALL
        ):
            leftover_comment = match_.group("leftover_comment").strip("\n")
            if leftover_comment:
                logger.info(f"leftover_comments: {leftover_comment}")
                leftover_comments.append(leftover_comment)
        return cls(
            leftover_comments=leftover_comments,
        )


class ExtractLeftoverComments(ChatGPT):
    def extract_leftover_comments(self, new_code, file_path, request, **kwargs):
        try:
            if not check_comments_presence(file_path, new_code):
                return []
            self.messages = [
                Message(role="system", content=system_message_prompt, key="system")
            ]
            response = self.chat(
                user_prompt.format(
                    new_code=new_code, file_path=file_path, request=request
                )
            )
            leftover_comments = LeftoverComments.from_string(response)
            return leftover_comments.leftover_comments
        except SystemExit:
            raise SystemExit
        except Exception as e:
            logger.error(f"An error occurred: {e}")
            return []


if __name__ == "__main__":
    leftover_comment_response = """<leftover_comments>
<leftover_comment>
TODO(sweep): We don't handle renamed files
</leftover_comment>
</leftover_comments>"""

    leftover_comments = LeftoverComments.from_string(leftover_comment_response)
    print(leftover_comments.leftover_comments)



from __future__ import annotations

import os
import re
import string
from dataclasses import dataclass
from difflib import unified_diff
from typing import Any, ClassVar, Literal, Type, TypeVar
from urllib.parse import quote

from loguru import logger
from pydantic import BaseModel, Field

from sweepai.utils.str_utils import (
    blockquote,
    clean_logs,
    create_collapsible,
    format_sandbox_success,
)

Self = TypeVar("Self", bound="RegexMatchableBaseModel")


class Message(BaseModel):
    role: (
        Literal["system"] | Literal["user"] | Literal["assistant"] | Literal["function"]
    )
    content: str | None = None
    name: str | None = None
    function_call: dict | None = None
    key: str | None = None

    @classmethod
    def from_tuple(cls, tup: tuple[str | None, str | None]) -> Self:
        if tup[0] is None:
            return cls(role="assistant", content=tup[1])
        else:
            return cls(role="user", content=tup[0])

    def to_openai(self) -> str:
        obj = {
            "role": self.role,
            "content": self.content,
        }
        if self.function_call:
            obj["function_call"] = self.function_call
        if self.role == "function":
            obj["name"] = self.name
        return obj


class Function(BaseModel):
    class Parameters(BaseModel):
        type: str = "object"
        properties: dict

    name: str
    description: str
    parameters: Parameters


class RegexMatchError(ValueError):
    pass


class RegexMatchableBaseModel(BaseModel):
    _regex: ClassVar[str]

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        # match = re.search(file_regex, string, re.DOTALL)
        match = re.search(cls._regex, string, re.DOTALL)
        if match is None:
            logger.warning(f"Did not match {string} with pattern {cls._regex}")
            raise RegexMatchError("Did not match")
        return cls(
            **{k: (v if v else "").strip("\n") for k, v in match.groupdict().items()},
            **kwargs,
        )


# todo (fix double colon regex): Update the split from "file_tree.py : desc" to "file_tree.py\tdesc"
# tab supremacy
def clean_filename(file_name: str):
    valid_chars = "-_./$[]%s%s" % (string.ascii_letters, string.digits)
    file_name = "".join(c for c in file_name if c in valid_chars)
    file_name = file_name.replace(" ", "")
    file_name = file_name.strip("`")
    return os.path.normpath(file_name)


def clean_instructions(instructions: str):
    return instructions.strip()


def create_error_logs(
    commit_url_display: str,
    sandbox_response: SandboxResponse,
    file_path: str = "",
):
    return (
        create_collapsible(
            f"Sandbox logs for {commit_url_display}",
            blockquote(
                "\n\n".join(
                    [
                        create_collapsible(
                            f"<code>{output}</code> {i + 1}/{len(sandbox_response.outputs)} {format_sandbox_success(sandbox_response.success)}",
                            f"<pre>{clean_logs(output)}</pre>",
                            i == len(sandbox_response.outputs) - 1,
                        )
                        for i, output in enumerate(sandbox_response.outputs)
                        if len(sandbox_response.outputs) > 0
                    ]
                )
            ),
            opened=True,
        )
        if sandbox_response
        else ""
    )


class ExtractionRequest(RegexMatchableBaseModel):
    use_tools: bool
    _regex = r"""<use_tools>\s+(?P<use_tools>.*?)</use_tools>"""

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        use_tools_pattern = r"""<use_tools>\s+(?P<use_tools>.*?)</use_tools>"""
        use_tools_match = re.search(use_tools_pattern, string, re.DOTALL)
        use_tools = use_tools_match.groupdict()["use_tools"].strip().lower() == "true"
        return cls(use_tools=use_tools)


class FileChangeRequest(RegexMatchableBaseModel):
    filename: str
    instructions: str
    change_type: (
        Literal["modify"]
        | Literal["create"]
        | Literal["delete"]
        | Literal["rename"]
        | Literal["rewrite"]
        | Literal["check"]
        | Literal["refactor"]
        | Literal["test"]
    )
    _regex = r"""<(?P<change_type>[a-z_]+)\s+file=\"(?P<filename>[a-zA-Z0-9/\\\.\[\]\(\)\_\+\- @\{\}]*?)\"( start_line=\"(?P<start_line>.*?)\")?( end_line=\"(?P<end_line>.*?)\")?( entity=\"(.*?)\")?( source_file=\"(?P<source_file>.*?)\")?( destination_module=\"(?P<destination_module>.*?)\")?( relevant_files=\"(?P<raw_relevant_files>.*?)\")?(.*?)>(?P<instructions>.*?)\s*<\/\1>"""
    is_completed: bool = False
    entity: str | None = None
    source_file: str | None = None
    old_content: str | None = None
    new_content: str | None = None
    raw_relevant_files: str | None = None
    # allow inf
    start_line: Any | int | str | None = None
    end_line: Any | int | str | None = None
    start_and_end_lines: list[tuple] = []
    comment_line: int | None = None
    sandbox_response: None = None
    failed_sandbox_test: bool | None = False
    parent: FileChangeRequest | None = None
    status: (
        Literal["succeeded"]
        | Literal["failed"]
        | Literal["queued"]
        | Literal["running"]
    ) = "queued"
    destination_module: str | None = None
    commit_hash_url: str | None = None

    def get_edit_url(self, repo_full_name: str, branch_name: str):
        url = f"https://github.com/{repo_full_name}/edit/{branch_name}/{self.filename}"
        if self.start_line and self.end_line:
            url += f"#L{self.start_line}-L{self.end_line}"
        return url

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        result = super().from_string(string, **kwargs)
        result.filename = result.filename.strip("/")
        result.instructions = result.instructions.replace("\n*", "\n•")
        if result.source_file:
            result.source_file = result.source_file.strip()
            if " " in result.source_file:
                result.source_file = result.source_file.split(" ")[0]
        if result.instructions.startswith("*"):
            result.instructions = "•" + result.instructions[1:]
        if result.start_line:
            try:
                result.start_line = int(result.start_line)
            except ValueError:
                result.start_line = None
        if result.end_line:
            try:
                result.end_line = int(result.end_line)
            except ValueError:
                result.start_line = None
        return result

    @property
    def relevant_files(self):
        if not self.raw_relevant_files:
            return []

        return [
            relevant_file
            for relevant_file in self.raw_relevant_files.split(" ")
            if relevant_file != self.filename
        ]

    @property
    def entity_display(self):
        if self.entity:
            return f"`{self.filename}:{self.entity}`"
        else:
            return f"`{self.filename}`"

    @property
    def status_display(self):
        if self.status == "succeeded":
            return "✓"
        elif self.status == "failed":
            if self.change_type == "modify":
                return "! No changes made"
            return "✗"
        elif self.status == "queued":
            return "▶"
        elif self.status == "running":
            return "⋯"
        else:
            raise ValueError(f"Unknown status {self.status}")

    @property
    def display_summary(self):
        if self.change_type == "check":
            return f"Running GitHub Actions for `{self.filename}`"
        return f"{self.change_type.capitalize()} `{self.filename}`"

    @property
    def summary(self):
        prefix = {"failed": "✗", "succeeded": "✓", "queued": "▶", "running": "⋯"}[
            self.status
        ] + " "
        return prefix + f"{self.change_type.capitalize()}\n{self.filename}"

    @property
    def color(self):
        color_map = {
            "failed": "red2",
            "succeeded": "#0ee832",
            "queued": "white",
            "running": "yellow",
        }
        return color_map[self.status]

    @property
    def entity_display_without_backtick(self):
        if self.entity:
            return f"{self.filename}:{self.entity}"
        else:
            return f"{self.filename}"

    @property
    def instructions_ticket_display(self):
        return self.instructions_display

    @property
    def instructions_display(self):
        # if self.change_type == "check":
        #     return f"Run GitHub Actions for `{self.filename}` with results:\n{self.instructions}"
        return f"{self.change_type.capitalize()} {self.filename} with contents:\n{self.instructions}"

    @property
    def diff_display(self):
        if self.old_content and self.new_content:
            diff = unified_diff(
                self.old_content.splitlines(keepends=True),
                self.new_content.splitlines(keepends=True),
            )
            diff_text = "".join(diff)
            return f"<pre>{diff_text}</pre>"
        return ""


class FileCreation(RegexMatchableBaseModel):
    commit_message: str
    code: str
    _regex = r"""<new_file(.*?)>(?P<code>.*)</new_file>"""
    # Regex updated to support ``` outside of <new_file> tags

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        re_match = re.search(cls._regex, string, re.DOTALL)

        if re_match is None:
            logger.info(f"Did not match {string} with pattern {cls._regex}")
            raise ValueError("No <new_file> tags or ``` found in code block")

        result = cls(
            code=re_match.groupdict()["code"].strip(),
            commit_message="Created file",
        )

        first_index = result.code.find("<new_file>")
        if first_index >= 0:
            last_index = result.code.rfind("</new_file>")
            result.code = result.code[first_index + len("<new_file>") : last_index]
        else:
            first_index = result.code.find("```")
            if first_index >= 0:
                last_index = result.code.rfind("```")
                file_extension = os.path.splitext(result.code)[1]
                if file_extension not in [".md", ".rst", ".mdx", ".txt"]:
                    result.code = result.code[first_index:last_index]

        result.code = result.code.strip()
        if result.code.endswith("</new_file>"):
            result.code = result.code[: -len("</new_file>")]
            result.code = result.code.strip()

        # Todo: Remove this?
        if len(result.code) == 1:
            result.code = result.code.replace("```", "")
            return result.code + "\n"

        if result.code.startswith("```"):
            first_newline = result.code.find("\n")
            result.code = result.code[first_newline + 1 :]

        result.code = result.code.strip()
        if result.code.endswith("```"):
            result.code = result.code[: -len("```")]
            result.code = result.code.strip()
        result.code += "\n"
        return result


class PullRequest(RegexMatchableBaseModel):
    title: str
    branch_name: str
    content: str
    _regex = r'''pr_title\s+=\s+"(?P<title>.*?)"\n+branch\s+=\s+"(?P<branch_name>.*?)"\n+pr_content\s+=\s+f?"""(?P<content>.*?)"""'''


class ProposedIssue(RegexMatchableBaseModel):
    title: str
    body: str
    issue_id: int | None = None
    _regex = r'<issue\s+title="(?P<title>.*?)">(?P<body>.*?)</issue>'


class Snippet(BaseModel):
    # pylint: disable=E1101
    """
    Start and end refer to line numbers
    """

    content: str = Field(repr=False)
    start: int
    end: int
    file_path: str

    def __eq__(self, other):
        if isinstance(other, Snippet):
            return (
                self.file_path == other.file_path
                and self.start == other.start
                and self.end == other.end
            )
        return False

    def __hash__(self):
        return hash((self.file_path, self.start, self.end))

    def get_snippet(self, add_ellipsis: bool = True, add_lines: bool = True):
        lines = self.content.splitlines()
        snippet = "\n".join(
            (f"{i + self.start}: {line}" if add_lines else line)
            for i, line in enumerate(lines[max(self.start - 1, 0) : self.end])
        )
        if add_ellipsis:
            if self.start > 1:
                snippet = "...\n" + snippet
            if self.end < self.content.count("\n") + 1:
                snippet = snippet + "\n..."
        return snippet

    def __add__(self, other):
        assert self.content == other.content
        assert self.file_path == other.file_path
        return Snippet(
            content=self.content,
            start=self.start,
            end=other.end,
            file_path=self.file_path,
        )

    def __xor__(self, other: "Snippet") -> bool:
        """
        Returns True if there is an overlap between two snippets.
        """
        if self.file_path != other.file_path:
            return False
        return self.file_path == other.file_path and (
            (self.start <= other.start and self.end >= other.start)
            or (other.start <= self.start and other.end >= self.start)
        )

    def __or__(self, other: "Snippet") -> "Snippet":
        assert self.file_path == other.file_path
        return Snippet(
            content=self.content,
            start=min(self.start, other.start),
            end=max(self.end, other.end),
            file_path=self.file_path,
        )

    @property
    def xml(self):
        return f"""<snippet source="{self.file_path}:{self.start}-{self.end}">\n{self.get_snippet()}\n</snippet>"""

    def get_xml(self, add_lines: bool = True):
        return f"""<snippet source="{self.file_path}:{self.start}-{self.end}">\n{self.get_snippet(add_lines)}\n</snippet>"""

    def get_url(self, repo_name: str, commit_id: str = "main"):
        num_lines = self.content.count("\n") + 1
        encoded_file_path = quote(self.file_path, safe="/")
        return f"https://github.com/{repo_name}/blob/{commit_id}/{encoded_file_path}#L{max(self.start, 1)}-L{min(self.end, num_lines)}"

    def get_markdown_link(self, repo_name: str, commit_id: str = "main"):
        num_lines = self.content.count("\n") + 1
        base = commit_id + "/" if commit_id != "main" else ""
        return f"[{base}{self.file_path}#L{max(self.start, 1)}-L{min(self.end, num_lines)}]({self.get_url(repo_name, commit_id)})"

    def get_slack_link(self, repo_name: str, commit_id: str = "main"):
        num_lines = self.content.count("\n") + 1
        base = commit_id + "/" if commit_id != "main" else ""
        return f"<{self.get_url(repo_name, commit_id)}|{base}{self.file_path}#L{max(self.start, 1)}-L{min(self.end, num_lines)}>"

    def get_preview(self, max_lines: int = 5):
        snippet = "\n".join(
            self.content.splitlines()[
                self.start : min(self.start + max_lines, self.end)
            ]
        )
        if self.start > 1:
            snippet = "\n" + snippet
        if self.end < self.content.count("\n") + 1 and self.end > max_lines:
            snippet = snippet + "\n"
        return snippet

    def expand(self, num_lines: int = 25):
        return Snippet(
            content=self.content,
            start=max(self.start - num_lines, 1),
            end=min(self.end + num_lines, self.content.count("\n") + 1),
            file_path=self.file_path,
        )

    @property
    def denotation(self):
        return f"{self.file_path}:{self.start}-{self.end}"


class DiffSummarization(RegexMatchableBaseModel):
    content: str
    _regex = r"""<file_summaries>(\n)?(?P<content>.*)$"""

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        result = super().from_string(string, **kwargs)
        result.content = result.content.replace("</file_summaries>", "", 1).strip()
        return cls(
            content=result.content,
        )


class PullRequestComment(RegexMatchableBaseModel):
    changes_required: str
    content: str
    _regex = r"""<changes_required>(?P<changes_required>.*)<\/changes_required>(\s+)<review_comment>(?P<content>.*)<\/review_comment>"""


class NoFilesException(Exception):
    def __init__(self, message="Sweep could not find any files to modify"):
        super().__init__(message)


class PRChangeRequest(BaseModel):
    params: dict


class MockPR(BaseModel):
    # Used to mock a PR object without creating a PR (branch will be created tho)
    file_count: int = 0  # Number of files changes
    title: str
    body: str
    pr_head: str
    base: Any
    head: Any
    assignee: Any = None

    id: int = -1
    state: str = "open"
    html_url: str = ""

    def create_review(self, *args, **kwargs):
        # Todo: used to prevent erroring in on_review.py file
        pass

    def create_issue_comment(self, *args, **kwargs):
        pass


@dataclass
class SandboxExecution:
    command: str
    output: str
    exit_code: int


class SandboxResponse(BaseModel):
    success: bool
    outputs: list[str]
    updated_content: str
    error_messages: list[str]


class MaxTokensExceeded(Exception):
    def __init__(self, filename):
        self.filename = filename


class UnneededEditError(Exception):
    def __init__(self, filename):
        self.filename = filename


class MatchingError(Exception):
    def __init__(self, filename):
        self.filename = filename


class EmptyRepository(Exception):
    def __init__(self):
        pass


@dataclass
class AssistantRaisedException(Exception):
    message: str



import json
import traceback
from datetime import datetime, timedelta
from threading import Thread
from typing import Any

import requests
from loguru import logger
from pydantic import BaseModel, Field
from pymongo import MongoClient

from sweepai.config.server import (
    DISCORD_LOW_PRIORITY_URL,
    DISCORD_MEDIUM_PRIORITY_URL,
    DISCORD_WEBHOOK_URL,
    GITHUB_BOT_USERNAME,
    IS_SELF_HOSTED,
    MONGODB_URI,
)
from sweepai.global_threads import global_threads

global_mongo_client = MongoClient(
    MONGODB_URI,
    serverSelectionTimeoutMS=20000,
    socketTimeoutMS=20000,
)


class ChatLogger(BaseModel):
    data: dict
    chat_collection: Any = None
    ticket_collection: Any = None
    _ticket_count_cache = {}
    _user_field_cache = {}
    expiration: datetime = None
    index: int = 0
    current_date: str = Field(
        default_factory=lambda: datetime.utcnow().strftime("%m/%Y/%d")
    )
    current_month: str = Field(
        default_factory=lambda: datetime.utcnow().strftime("%m/%Y")
    )
    active: bool = False  # refers to whether it was an auto-created PR or if it was created by the user with intent

    def __init__(self, data: dict[str, str] = {}, mock=False, **kwargs):
        super().__init__(data=data, **kwargs)  # Call the BaseModel's __init__ method
        key = MONGODB_URI
        if key is None:
            logger.warning("Chat history logger has no key")
            return
        if not mock:
            try:
                client = global_mongo_client
                db = client["llm"]
                self.chat_collection = db["chat_history"]
                self.ticket_collection = db["tickets"]
                # For 'username' index on ticket_collection
                self.ticket_collection.index_information()
                self.ticket_collection.create_index("username")

                # For 'expiration' index on chat_collection
                self.chat_collection.index_information()
                self.chat_collection.create_index(
                    "expiration", expireAfterSeconds=2419200
                )
                self.expiration = datetime.utcnow() + timedelta(days=1)
            except SystemExit:
                raise SystemExit
            except Exception as e:
                logger.warning("Chat history could not connect to MongoDB")
                logger.warning(e)

    def _add_chat(self, additional_data):
        if self.chat_collection is None:
            logger.error("Chat collection is not initialized")
            return
        document = {
            **self.data,
            **additional_data,
            "expiration": self.expiration,
            "index": self.index,
        }
        self.index += 1
        self.chat_collection.insert_one(document)

    def add_chat(self, additional_data):
        thread = Thread(target=self._add_chat, args=(additional_data,))
        thread.start()
        global_threads.append(thread)

    def _add_successful_ticket(self, gpt3=False):
        if self.ticket_collection is None:
            logger.error("Ticket Collection Does Not Exist")
            return

        username = self.data.get("assignee", self.data["username"])
        update_fields = {self.current_month: 1, self.current_date: 1}

        if gpt3:
            key = f"{self.current_month}_gpt3"
            update_fields = {key: 1}

        self.ticket_collection.update_one(
            {"username": username}, {"$inc": update_fields}, upsert=True
        )

        ticket_count = self.get_ticket_count()
        should_decrement = (self.is_paying_user() and ticket_count >= 500) or (
            self.is_consumer_tier() and ticket_count >= 20
        )

        if should_decrement:
            self.ticket_collection.update_one(
                {"username": username}, {"$inc": {"purchased_tickets": -1}}, upsert=True
            )

        logger.info(f"Added Successful Ticket for {username}")

    def add_successful_ticket(self, gpt3=False):
        thread = Thread(target=self._add_successful_ticket, args=(gpt3,))
        thread.start()
        global_threads.append(thread)

    def _cache_key(self, username, field, metadata=""):
        return f"{username}_{field}_{metadata}"

    def get_ticket_count(
        self, use_date: bool = False, gpt3: bool = False, purchased: bool = False
    ) -> int:
        if self.ticket_collection is None:
            logger.error("Ticket Collection Does Not Exist")
            return
        username = self.data["username"]
        cache_key = self._cache_key(
            username, "ticket_count", f"{use_date}_{gpt3}_{purchased}"
        )

        if cache_key in self._ticket_count_cache:
            return self._ticket_count_cache[cache_key]
        tracking_date = self.current_date if use_date else self.current_month
        if gpt3:
            tracking_date = f"{self.current_month}_gpt3"
        query = {"username": username}
        result = self.ticket_collection.find_one(query, {tracking_date: 1, "_id": 0})
        if purchased:
            ticket_count = result.get("purchased_tickets", 0) if result else 0
        else:
            ticket_count = result.get(tracking_date, 0) if result else 0
        self._ticket_count_cache[cache_key] = ticket_count
        return ticket_count

    def _get_user_field(self, field: str):
        if self.ticket_collection is None:
            logger.error("Ticket Collection Does Not Exist")
            return None

        username = self.data["username"]
        cache_key = self._cache_key(username, field)

        if cache_key in self._user_field_cache:
            return self._user_field_cache[cache_key]

        result = self.ticket_collection.find_one({"username": username}, {field: 1})

        user_field_value = result.get(field, False) if result else False
        self._user_field_cache[cache_key] = user_field_value

        return user_field_value

    def is_consumer_tier(self):
        if IS_SELF_HOSTED:
            return True
        return self._get_user_field("is_trial_user")

    def is_paying_user(self):
        if IS_SELF_HOSTED:
            return True
        return self._get_user_field("is_paying_user")

    def use_faster_model(self):
        if IS_SELF_HOSTED:
            return False
        if self.ticket_collection is None:
            logger.error("Ticket Collection Does Not Exist")
            return True
        purchased_tickets = self.get_ticket_count(purchased=True)
        if self.is_paying_user():
            return self.get_ticket_count() >= 500 and purchased_tickets == 0
        if self.is_consumer_tier():
            return self.get_ticket_count() >= 20 and purchased_tickets == 0
        return (
            (self.get_ticket_count() >= 5 or self.get_ticket_count(use_date=True) > 3)
            and purchased_tickets == 0
        ) or not self.active


def discord_log_error(content, priority=0):
    """
    priority: 0 (high), 1 (medium), 2 (low)
    """
    logger.error(content)
    if GITHUB_BOT_USERNAME != "sweep-ai[bot]":  # disable for dev
        return
    try:
        url = DISCORD_WEBHOOK_URL
        if priority == 1:
            url = DISCORD_MEDIUM_PRIORITY_URL
        if priority == 2:
            url = DISCORD_LOW_PRIORITY_URL

        data = {
            "content": f"Traceback:\n\n{traceback.format_exc()}\n\nMessage:\n\n```\n{content}\n```"
        }
        headers = {"Content-Type": "application/json"}
        requests.post(url, data=json.dumps(data), headers=headers)
        # Success: response.status_code == 204:
    except SystemExit:
        raise SystemExit
    except Exception as e:
        logger.error(f"Could not log to Discord: {e}")


if __name__ == "__main__":
    chat_logger = ChatLogger(
        {
            "username": "kevinlu1248",
        }
    )
    print(chat_logger.get_ticket_count())
